import requests

from common.env_config import env
from common.api.presigned_url import (
    get_upload_presigned_url,
    get_download_presigned_url
)
from common.exceptions import RemoteCallException, DecoratorException
from ..logging_config import logger



class S3Uploader():
    """This class is responsible for handling file transfer 
       to and from Amazon S3 trough presigned URLs."""

    @staticmethod
    def upload_bytes(bytes, job_id, path):
        presigned_resp = get_upload_presigned_url(job_id, path)
        
        filename = path.split('/')[-1]
        # Create a file-like object from bytes
        files = {'file': (filename, bytes)}
        response = requests.post(presigned_resp.get('url'), data=presigned_resp.get('fields'), files=files) 
        if response.status_code != 204:
            raise DecoratorException(
                f'Upload failed with http code: {response.status_code} \n {response.text}')
            

    @staticmethod
    def upload_file(file_path, job_id, path):
        presigned_resp = get_upload_presigned_url(job_id, path)

        with open(file_path, 'rb') as f:
            files = {'file': (path, f)}
            response = requests.post(presigned_resp.get('url'), data=presigned_resp.get('fields'), files=files)
            logger.debug(response)
            if response.status_code != 204:
                raise DecoratorException(
                    f'Upload failed with http code: {response.status_code} \n {response.text}')

class S3Downloader():
    """This class is responsible for handling file transfer 
       to and from Amazon S3 trough presigned URLs."""
    @staticmethod
    def download_file_to_bytes(job_id, file_path):
        presigned_resp = get_download_presigned_url(job_id, file_path)
        resp = requests.get(presigned_resp.get('url'))
        if resp.status_code != 200:
            return None
        else:
            return resp.content


   


import typer
import csv
from datetime import datetime, timezone
from typing import Optional
from typing_extensions import Annotated
from common.api.list_jobs import list_jobs
from common.observability.main import track_command_usage
from common.console import console
from cli.cmds.login import login

@track_command_usage("get")
def get(
  type: Annotated[Optional[str], typer.Argument(help='runs')] = None,
  all: Annotated[bool, typer.Option("--all", help="Show runs in the running state")] = None,
  verbose: Annotated[bool, typer.Option("--v", help="Show more info about the run")] = None,
  save: Annotated[str, typer.Option("--save", help="Output the results as a .csv to <file> instead of printing")] = None
):
    """
    Get a list of "running' runs
    """

    verbose_data = [
        ["ID", "Name", "Status", "Duration (HH:MM:SS)", "Start Time (local)", "End time (local)", "Rate", "Tokens Used", "UUID"],
        ["--", "----", "------", "-------------------", "------------------", "----------------", "----", "-----------", "----"],
    ]

    simple_data = [
        ["ID", "Name", "Status", "Rate", "Tokens Used" ],
        ["--", "----", "------", "----", "-----------" ],
    ]
    
    '''
    console.print(f"===== get =====")
    console.print(f"type: {type}")
    console.print(f"all: {all}")
    console.print(f"verbose: {verbose}")
    console.print(f"save: {save}")
    '''

    login()

    if type == "runs":
        try:
            jobs = list_jobs()
            if len(jobs) == 0:
                console.print("No runs found")
                return

            for key, val in jobs.items():
                if not is_running(val["status"]) and not all:
                    continue
                duration = convert_ms(val['durationMs']) if val['durationMs'] != None else None
                local_start_time = convert_utc_to_local(val['startDate']) if val['startDate'] != None else None
                local_end_time = convert_utc_to_local(val['endDate']) if val['endDate'] != None else None

                if verbose:
                    verbose_data.append([key, val['name'], val['status'], duration, local_start_time, local_end_time, val['startTokenRate'], val['tokensUsed'], val['id']])    
                else:   
                    simple_data.append([key, val['name'], val['status'], val['startTokenRate'], val['tokensUsed']])

            if save:
                # we are writing to a file
                if verbose:
                    if not save.endswith(".csv"):
                        save += ".csv"
                    with open(save, "w", newline="") as file:
                        try:
                            writer = csv.writer(file)
                            writer.writerows(verbose_data)
                            console.print(f"[green]Successfully wrote file:{save}[/green]")
                        except Exception as e:
                            console.print(f"[red]Error writing to file:{save} error: {e}[/red]")
                else:
                    if not save.endswith(".csv"):
                        save += ".csv"
                    with open(save, "w", newline="") as file:
                        try:
                            writer = csv.writer(file)
                            writer.writerows(simple_data)
                            console.print(f"[green]Successfully wrote file:{save}[/green]")
                        except Exception as e:
                            console.print(f"[red]Error writing to file:{save} error: {e}[/red]")
            else:
                # we are printing
                if verbose:
                    col_widths = [max(len(str(row[i])) for row in verbose_data) for i in range(len(verbose_data[0]))]
                    for row in verbose_data:
                        console.print("  ".join(f"{str(item):<{width}}" for item, width in zip(row, col_widths)))
                else:
                    col_widths = [max(len(str(row[i])) for row in simple_data) for i in range(len(simple_data[0]))]
                    for row in simple_data:
                        console.print("  ".join(f"{str(item):<{width}}" for item, width in zip(row, col_widths)))
        except Exception as e:
            console.print(f"[red]Error getting runs: {e}[/red]")
    else:    
        console.print("[red]Please provide a valid type, like 'runs'[/red]")

def convert_ms(ms):
    """Converts milliseconds to hours, minutes, and seconds."""
    seconds, ms = divmod(ms, 1000)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def convert_utc_to_local(utc_time_str):
    """Converts a UTC time string to a local time string."""
    # Parse the UTC time string to a datetime object
    utc_time = datetime.strptime(utc_time_str, "%Y-%m-%dT%H:%M:%S.%fZ")

    # Set the datetime object to UTC timezone
    utc_time = utc_time.replace(tzinfo=timezone.utc)

    # Convert UTC time to local time
    local_time = utc_time.astimezone()

    # Format the local time as a string
    local_time_str = local_time.strftime("%Y-%m-%d %H:%M:%S")

    return local_time_str


def is_running(status):
    """Check if the status is running"""
    if (status != "terminated" 
        and status != "terminate_job"
        and status != "complete" 
        and status != "failed" 
        and status != "execution_error"):
        return True
    else:
        return False
           


import zipfile
import os
import shutil
import fnmatch
from dataclasses import dataclass
from typing import List
from common.constants import FILE_SIZE_THRESHOLD
from positron_job_runner.runner_env import runner_env
from positron_job_runner.cloud_storage import cloud_storage
from positron_job_runner.cloud_logger import logger


EXCLUDED_DIRECTORIES = ['venv', '.venv', '.git', '__pycache__', 'job-execution', '.robbie', '.ipynb_checkpoints', 'persistent-disk']
S3_BASE_PATH = f"{runner_env.JOB_OWNER_EMAIL}/{runner_env.JOB_ID}"
S3_RESULT_PATH = f"{S3_BASE_PATH}/result"

def download_workspace_from_s3():
    """Download the workspace from S3"""
    s3_key = f"{S3_BASE_PATH}/workspace.zip"
    local_zip_path = os.path.join(runner_env.RUNNER_CWD, 'workspace.zip')
    cloud_storage.download_file(s3_key, local_zip_path)
    # the file may or may not exist

def copy_workspace_to_job_execution():
    """Copies the workspace from job-controller to job-execution"""
    local_zip_path = os.path.join(runner_env.RUNNER_CWD, 'workspace.zip')
    if os.path.exists(local_zip_path):
        destination_zip_path = os.path.join(runner_env.JOB_CWD, 'workspace.zip')
        print(f"Copying: {local_zip_path} to {destination_zip_path}")
        shutil.copy(local_zip_path, destination_zip_path)
        logger.debug(f"Copied workspace.zip to {destination_zip_path}")
    else:
        logger.error("Workspace zip not found")

def upload_results_to_s3():
    """Uploads the results to S3"""
    try:
        logger.info('Copying results to cloud storage...')

        results_dir = runner_env.JOB_CWD
        os.makedirs(results_dir, exist_ok=True)

        result_files = get_file_paths(path=results_dir, excluded_dirs=EXCLUDED_DIRECTORIES)

        ignore_name_patterns = []
        # read in the ignore file
        for entry in result_files:
            if ".robbieignore" == entry.name:
                with open(entry.full_path, "r") as f:
                    ignore_name_patterns = f.read().splitlines()

        # Create a zip of the result directory
        results_zip_file_name = f"{results_dir}/result.zip"
        with zipfile.ZipFile(results_zip_file_name, "w", zipfile.ZIP_DEFLATED) as zipf:
            for file in result_files:
                ignore = False
                for pattern in ignore_name_patterns:
                    if fnmatch.fnmatch(file.name, pattern):
                        logger.debug(f"Ignoring: {file.name}")
                        ignore = True
                        break
                if not ignore:
                    logger.debug(f"Adding to zip: {file.name}")
                    zipf.write(file.full_path, arcname=file.name)

        file_size = os.path.getsize(results_zip_file_name)
        if (file_size >= FILE_SIZE_THRESHOLD):
            size_in_mb = round(file_size / (1024 * 1024), 2)
            logger.warning(f"Results Archive Size: {size_in_mb} Mb. It might take a long time to upload it.")

        logger.debug(f"Uploading to cloud storage: {results_zip_file_name}")
        cloud_storage.upload_file(results_zip_file_name, f"{S3_RESULT_PATH}/result.zip")

        # Upload raw files to S3
        for file in result_files:
            logger.debug(f"Uploading to cloud storage: {file.name}")
            s3_key = f"{S3_RESULT_PATH}/{file.name}"
            cloud_storage.upload_file(file.full_path, s3_key)

        logger.info('Results uploaded to S3 successfully')

    except Exception as e:
        logger.error(f"Failed to upload results to S3: {e}")


@dataclass
class FileEntry:
    full_path: str
    name: str
    size: int

def get_file_paths(path: str, excluded_dirs: List[str], excluded_files: List[str] = []) -> List[FileEntry]:
    """should return a list relative paths that should included in the results"""
    all_files: List[FileEntry] = []
    for root, dirs, files in os.walk(path):
        for exclude_dir in excluded_dirs:
            try:
                dirs.remove(exclude_dir)
            except ValueError:
                pass
        for file in files:
            if file in excluded_files:
                continue
            full_path = os.path.join(root, file)
            file_size = os.path.getsize(full_path)
            rel_path = os.path.relpath(full_path, path)
            all_files.append(FileEntry(name=rel_path, size=file_size, full_path=full_path))
    return all_files





import os
import typer
import pyfiglet
from typing_extensions import Annotated
from common.console import console
from common.constants import JOB_CONF_YAML_PATH
from common.observability.main import track_command_usage
from cli.interactive import (
    prompt_and_build_positron_job_config,
    get_job_config_yaml_name
)
from common.console import console, ROBBIE_BLUE, ROBBIE_DEFAULT


@track_command_usage("configure")
def config() -> None:
    """
    Build a Robbie job configure file (job_config.yaml) interactively.

    """
    ascii_banner = pyfiglet.figlet_format("Robbie")
    console.print(ascii_banner, style='#41a7ff')

    console.print("[green]✔[/green] Current working directory: ", os.getcwd(), style=ROBBIE_BLUE)

    console.print(f"\nPlease follow the prompts to build a job configration .yaml file ([{ROBBIE_DEFAULT}][] = default[/{ROBBIE_DEFAULT}], <tab> for menu, contol-c to exit):", style=ROBBIE_BLUE)
    
    cfg = prompt_and_build_positron_job_config(
        job_config_yaml = None,
        cmd_runner_dash_i=False,
    )

    if cfg:
        filename = get_job_config_yaml_name()
        if filename:
            cfg.write_to_file(filename=filename)
            console.print(f"[green]Successfully wrote config file {JOB_CONF_YAML_PATH}'")
            console.print("File contents:")
            with open(filename, 'r') as f:
                print(f.read())





    



import requests
import json
import pprint
import re
from typing import Optional, List, Dict, Any
from pydantic import BaseModel, Field
from common.utils import undefined
from common.config import PositronJob
from common.cli_args import args as cli_args
from common.env_config import env
from common.exceptions import RemoteCallException, RobbieException
from common.logging_config import logger
from common.env_defaults import current

class GetEnvRecBody(BaseModel):
    """
    Maps to the request body of the Create Job API
    """
    model: Optional[str] = Field(default=undefined)
    prompt: List[List[str]] = Field(default=undefined)

    def http_dict(self) -> Dict[str, Any]:
        """
        Enables dropping fields that were never set and should be treated as undefined
        """
        return {k: v for k, v in self.__dict__.items() if v is not undefined}

def get_env_rec( 
    system_prompt: str,
    human_prompt: str,
    model: str = None
) -> str:
    """
        Use GenAI to get an environment recommendation

        Args:
            python_code (str): The python code to be run

        Returns:
            str: The environment_id recommendation
    """
    prompt = [["system", system_prompt], ["human", human_prompt]]

    if model is None:
        data = GetEnvRecBody(
            prompt=prompt
        )
    else: 
        data = GetEnvRecBody(
            model=model,
            prompt=prompt
        )
    # print(pprint.pformat(data.http_dict()))

    
    url = f'{env.API_BASE}/get-env-rec'
    logger.debug(f'Calling: {url}')
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN}
    response = requests.post(url, headers=Headers, json=data.http_dict())
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        # TODO: need standard error codes in errors
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'get_env_rec() failed with http code: {response.status_code} \n {response.text}')
    else:
        # logger.debug(response.json())
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()



from common.api.funding_envs_images import *

# Class and singleton builds a list of tuples from the DB results
class FundingSources: 
    is_init: bool = False
    my_fs: dict

    def __init__(self, fs_arg: dict):
        if self.is_init:
            raise ValueError('FundingSources.load() already initialized')
        else:
            self.init = True
            self.my_fs= fs_arg

    @staticmethod
    def load():
        fs = list_funding_sources()
        if len(fs) == 0:
            return None
        # Loop through and add a custom "menu" item to each dict 
        for key, val in fs.items(): 
                val[FS_MENU] = f'{val[FS_NAME]} ({val[FS_TOKENS]} tokens available)'
        return FundingSources(fs)
        
    # Prompt toolkit needs a list of strings to display in the menu 
    def menu_items(self) -> list: 
        ret_list: list = []
        for _, val in self.my_fs.items():
            # just show names
            ret_list.append(val[FS_MENU])
        return ret_list
    
    def auto_complete_items(self) -> list: 
        ret_list: list = []
        for _, val in self.my_fs.items():
            # just show names
            ret_list.append((val[FS_ID],val[FS_MENU]))
        return ret_list

    # Return 'funding_group_id' using the val returned from session.prompt() 
    def id_from_menu_item(self, menu_item: str) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_ID)
        return None

    def default_env_id_from_menu_item(self, menu_item: str) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_env_name_from_menu_item(self, menu_item: str) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_DEF_ENV_NAME)
        return None
    
    def default_env_name_from_fs_id(self, id: str) -> str:
        for _, val in self.my_fs.items():
                if (val[FS_ID] == id):
                    return val.get(FS_DEF_ENV_NAME)
        return None
    
    def default_env_id_from_fs_id(self, id: str) -> str:
        for _, val in self.my_fs.items():
                if (val[FS_ID] == id):
                    return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_env_id(self) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_funding_source_id(self) -> str: 
        for _, val in self.my_fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return val.get(FS_ID)
        return None
    
    def default_image_name(self, id: str) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_ID] == id):
                return val.get(FS_DEFAULT_IMAGE_NAME)
        return None
    
    def default_image_id(self, id: str) -> str:
        for _, val in self.my_fs.items():
            if (val[FS_ID] == id):
                return val.get(FS_DEFAULT_IMAGE_ID)
        return None


# singleton for Environments
class Environments: 
    is_init: bool = False
    my_envs: dict

    def __init__(self, env_arg):
         if self.is_init:
            raise ValueError('Environments.load() already initialized')
         else:
            self.my_envs = env_arg
            self.is_init = True

    @staticmethod
    def load(fs_id: str):
        return_dict = {}
        envs = list_environments(fs_id)
        if len(envs) == 0:
            return None
        for key, val in envs.items():
            if not val.get(ENV_DELETED):
                return_dict[key] = val
                val[ENV_MENU_ITEM] = f"{val['environmentName']} ({val['tokensPerHour']} Tokens/Hour)" # shows in menu

        return Environments(return_dict)

    def menu_items(self) -> list: 
        menu_list = []
        for _, val in self.my_envs.items():
            menu_list.append(val[ENV_MENU_ITEM])
        return menu_list
    
    def auto_complete_items(self) -> list: 
        ret_list: list = []
        for _, val in self.my_envs.items():
            ret_list.append((val[ENV_ID],val[ENV_MENU_ITEM]))
        return ret_list

    def id_from_menu_item(self, menu_item: str) -> str:
        for _, val in self.my_envs.items():
            if (val[ENV_MENU_ITEM] == menu_item):
                return val.get(ENV_ID)
        return None

    def tokens_per_hour(self, env_id: str) -> str:
        for _, val in self.my_envs.items():
            if (val[ENV_ID] == env_id):
                return val.get(ENV_TPH)
        return None
    
    def cluster_type_from_env_id(self, env_id: str) -> str:
        for _, val in self.my_envs.items():
            if (val[ENV_ID] == env_id):
                return val.get(ENV_CLUSTER_TYPE)
        return None
        


# singleton for Environments
class Images: 
    is_init: bool = False
    my_images: dict

    def __init__(self, image_arg):
         if self.is_init:
            raise ValueError('Images.load() already initialized')
         else:
            self.my_images = image_arg
            self.is_init = True

    @staticmethod
    def load(fs_id: str, env_id: str):
        images = list_images(fs_id, env_id)
        if len(images) == 0:
            return None
        for _, val in images.items():
            val[IMAGE_MENU_ITEM] = f"{val[IMAGE_NAME]}" # shows in menu
        return Images(images)

    def menu_items(self) -> list: 
        menu_list = []
        for _, val in self.my_images.items():
            if val.get(IMAGE_DELETED) == False:
                menu_list.append(val[IMAGE_MENU_ITEM])
        return menu_list
    
    def auto_complete_items(self) -> list:
        menu_list = []
        for _, val in self.my_images.items():
            if val.get(IMAGE_DELETED) == False:
                menu_list.append(val[IMAGE_MENU_ITEM])
        return menu_list
        

    def name_from_menu_item(self, menu_item: str) -> str:
        for _, val in self.my_images.items():
            if (val[IMAGE_MENU_ITEM] == menu_item):
                return val.get(IMAGE_NAME)
        return None
    


import typer
import pyfiglet
from typing import Optional
from typing_extensions import Annotated
from common.common_dump import common_dump
from common.observability.main import track_command_usage
from common.env_defaults import current
from common.console import console

# robbie "dump" command
# @track_command_usage("dump")
def dump(
    machine: Annotated[bool, typer.Option("--machine", help="Machine related info on the local machine")] = None,
    conda: Annotated[bool, typer.Option("--conda", help='Conda related parameters on the local machine')] = None,
    python: Annotated[bool, typer.Option("--python", help="Python related info on the local machine")] = None,
    sdk: Annotated[bool, typer.Option("--sdk", help="Robbie SDK related configuration")] = None,
    job_config: Annotated[bool, typer.Option("--job_config", help="Job related configuration")] = None,
    job_config_name: Annotated[Optional[str], typer.Argument()] = None,
    save: Annotated[str, typer.Option("--save", help="Saves configuration info to a file that can be emailed.")] = None 
):

    ascii_banner = pyfiglet.figlet_format(f"Robbie - {current.name}")
    console.print(ascii_banner, style='#41a7ff')

    common_dump(
        machine,
        conda,
        python,
        sdk,
        (True if job_config else False),
        (job_config_name if job_config_name else None),
        save
    )



import uuid
import typer
from rich.prompt import Confirm
from typing import Optional, List
from typing_extensions import Annotated
from common.api.list_jobs import list_jobs
from common.api.terminate_job import terminate_job
from common.api.get_job import get_job
from common.observability.main import track_command_usage
from common.console import console
from common.utils import is_valid_uuid
from cli.cmds.login import login


@track_command_usage("kill")
def kill(
  type: Annotated[str, typer.Argument(help='run')],
  ids: Annotated[List[str], typer.Argument(help="ID(s) of the run to kill")] = None,
  all: Annotated[bool, typer.Option("--all", help="Kills all runs")] = None,
  yes: Annotated[bool, typer.Option("--y", help="Skip confirmation")] = None
):
    """
    Kill a run

    Usage: robbie kill run <id> | --all

    `id` can be the positional ID returned from get runs or the UUID of the run

    """

    if type != "run":
        console.print("Please provide a valid object type to kill")
        return
    
    if ids is not None and all:
        console.print("Please provide a run ID or use --all to kill all runs")
        return
    
    login()

    # get all the jobs
    try:
        jobs = list_jobs()
        if len(jobs) == 0:
            console.print("No jobs found")
            return

        if all:
            # loop through and kill all running jobs
            if yes or Confirm.ask("[bold red]Are you sure you want to kill all runs?", default=False):
                for key, val in jobs.items():
                    if is_running(val["status"]):
                        try:
                            terminate_job(val["id"], "Killed by user CLI command")
                            console.print(f"Successfully killed run: {key}, name: {val['name']} - uuid: {val['id']}")
                        except Exception as e:
                            console.print(f"Failed to kill run: {key}, name: {val['name']} - uuid: {val['id']}, error: {e}")
            else:
                console.print("No runs were killed")
            return

        # Now figure out which type of ID we are dealing with

        # Support the ability to kill by UUID too
        for id in ids:
            if is_valid_uuid(id):
                if yes or Confirm.ask(f"[bold red]Are you sure you want to kill run:{id}?", default=False):
                    try:
                        val = get_job(id)
                        if is_running(val["status"]):
                            terminate_job(id, "Killed by using kill CLI command")
                            console.print(f"Successfully killed run: {id}, name: {val['name']} - uuid: {val['id']}")
                        else:
                            console.print(f"Sorry, run not runnning: {id}, name: {val['name']} - status: {val['status']}")
                    except Exception as e:
                        console.print(f"Failed to kill run: {id}, error: {e}")
                else:
                    console.print("Run not killed")

            # User provided a short run ID from the `robbie get runs` command
            for key, val in jobs.items():
                if key == id:
                    if yes or Confirm.ask(f"[bold red]Are you sure you want to kill run:{id}?", default=False):
                        if is_running(val["status"]):
                            try:
                                terminate_job(val["id"], "Killed by using kill CLI command")
                                console.print(f"Successfully killed run: {id}, name: {val['name']} - uuid: {val['id']}")
                            except Exception as e:
                                console.print(f"Failed to kill run: {id}, name: {val['name']} - uuid: {val['id']}, error: {e}")
                        else:
                            console.print(f"Sorry, run not runnning: {id}, name: {val['name']} - uuid: {val['id']} - status: {val['status']}")
                    else:
                        console.print("Run not killed")
        return
    except Exception as e:
        console.print(f"Failed to list jobs: {e}")
        return


def is_running(status):
    """Check if the status is running"""
    if (status != "terminated" 
        and status != "terminate_job"
        and status != "complete" 
        and status != "failed" 
        and status != "execution_error"):
        return True
    else:
        return False
    


# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import absolute_import

import fnmatch
import os
import shutil
from typing import List, Optional, Callable, Union
from common.config import PositronJob

class CustomFileFilter:
    """Configuration that specifies how the local working directory should be packaged."""

    def __init__(self, *, ignore_name_patterns: List[str] = None):
        """Initialize a CustomFileFilter.

        Args:
            ignore_name_patterns (List[str]): ignore files or directories with names
              that match one of the glob-style patterns. Defaults to None.
        """

        if ignore_name_patterns is None:
            ignore_name_patterns = []

        self._workdir = os.getcwd()
        self._ignore_name_patterns = ignore_name_patterns

    @property
    def ignore_name_patterns(self):
        """Get the ignore name patterns."""
        return self._ignore_name_patterns

    @property
    def workdir(self):
        """Get the working directory."""
        return self._workdir



def copy_workdir(
    dst: str,
    custom_file_filter: Optional[CustomFileFilter] = None,
    print_tree: bool = True,
):
    """Copy the local working directory to the destination.

    Args:
        dst (str): destination path.
        custom_file_filter (CustomFileFilter): configuration that
            specifies how the local working directory should be packaged.
    """
    # so we can print out what is shipped over
    from common.print_banner import workspace_file_tree

    """
    def _ignore_patterns(path: str, names: List):  # pylint: disable=unused-argument
        ignored_names = set()
        if custom_file_filter.ignore_name_patterns is not None:
            for pattern in custom_file_filter.ignore_name_patterns:
                ignored_names.update(fnmatch.filter(names, pattern))
        return ignored_names
    """

    def _ignore_patterns(path: str, names: List): 
        to_ignore = []
        for name in names:
            full_path = os.path.join(path, name)

            "when `default` is set in the config file"
            if os.path.isfile(full_path):
                if (name == ".python-version" or
                    name == ".DS_Store"
                ):
                    to_ignore.append(name)
                    continue

            if os.path.isdir(full_path):
                if (name == "__pycache__" or 
                    name == "job-execution" or 
                    name == ".robbie" or 
                    name == ".ipynb_checkpoints" or
                    name == "venv" or 
                    name == ".pyenv" or 
                    name == ".git" or
                    name == ".venv"
                ):
                    to_ignore.append(name)
                    continue

            ignoring = False
            # go through each pattern if they exist
            if (custom_file_filter is not None and 
                custom_file_filter.ignore_name_patterns is not None
            ):
                for pattern in custom_file_filter.ignore_name_patterns:
                    if fnmatch.fnmatch(name, pattern):
                        to_ignore.append(name)
                        ignoring = True
                        break
            # if not ignoring:
            #     workspace_file_tree.add(f"[yellow]{full_path}, size: {os.path.getsize(full_path)} bytes[/yellow]")
            
        return to_ignore

    def fake_copytree(src, ignore=None):
        """
        Recursively prints the tree from `src` using the ignore function.
        Calls the `ignore` function for each directory visited.

        Args:
            src (str): Source directory path.
            dst (str): Destination directory path.
            ignore (callable, optional): A callable that takes the directory path and its contents 
                                         (list of names) and returns a list of names to ignore.
        """
        # Ensure the destination directory exists
        # os.makedirs(dst, exist_ok=True)

        # Get the list of items in the source directory
        entries = os.listdir(src)

        # Apply the ignore function if provided
        ignored_names = ignore(src, entries) if ignore else []

        for entry in entries:
            full_path = os.path.join(src, entry)

            if entry in ignored_names:
                continue  # Skip ignored items

            src_path = os.path.join(src, entry)

            if os.path.isdir(src_path):
                # Recursively copy subdirectory
                fake_copytree(src_path, ignore)
            else:
                # Add file to tree
                workspace_file_tree.add(f"[yellow]{os.path.relpath(full_path)}, size: {os.path.getsize(full_path)} bytes[/yellow]")

    # start of execution
    _ignore = None
    _src = os.getcwd()
    if not custom_file_filter:
        _ignore = _ignore_patterns
    elif isinstance(custom_file_filter, CustomFileFilter):
        _ignore = _ignore_patterns
        _src = custom_file_filter.workdir

    if print_tree:
        fake_copytree(_src, _ignore)
    else:
        shutil.copytree(
            _src,
            dst,
            ignore=_ignore,
            dirs_exist_ok=True,
        )








# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""An entry point for runtime environment. This must be kept independent of SageMaker PySDK"""
from __future__ import absolute_import

import argparse
import getpass
import sys
import os
import shutil
import pathlib

if (os.getenv('POSITRON_CLOUD_ENVIRONMENT')):
    from positron_job_runner.cloud_logger import logger
else:
    from common.logging_config import logger

from positron_job_runner.runtime_environment_manager import (
        RuntimeEnvironmentManager,
        _DependencySettings,
        _run_shell_cmd
)
from positron_job_runner.runner_env import runner_env

SUCCESS_EXIT_CODE = 0
DEFAULT_FAILURE_CODE = 1

REMOTE_FUNCTION_WORKSPACE = "sm_rf_user_ws"
BASE_CHANNEL_PATH = "/opt/ml/input/data"
FAILURE_REASON_PATH = "/opt/ml/output/failure"
JOB_OUTPUT_DIRS = ["/opt/ml/output", "/opt/ml/model", "/tmp"]
PRE_EXECUTION_SCRIPT_NAME = "pre_exec.sh"
JOB_REMOTE_FUNCTION_WORKSPACE = "workspace"
SCRIPT_AND_DEPENDENCIES_CHANNEL_NAME = "pre_exec_script_and_dependencies"


def _bootstrap_runtime_env(
    client_python_version: str,
    conda_env: str = None,
    dependency_settings: _DependencySettings = None,
    pre_exec_commands: str = None,
):
    """Bootstrap runtime environment for remote function invocation.

    Args:
        client_python_version (str): Python version at the client side.
        conda_env (str): conda environment to be activated. Default is None.
        dependency_settings (dict): Settings for installing dependencies.
    """

    workspace_unpack_dir = _unpack_user_workspace()
    if not workspace_unpack_dir:
        logger.info("No workspace to unpack and setup.")
        return

    print("workspace_unpack_dir: ", workspace_unpack_dir)
    
    # FUTURE Use
    #_handle_pre_exec_commands(pre_exec_commands)

    _install_dependencies(
        dependency_file_dir = workspace_unpack_dir,
        conda_env = conda_env,
        client_python_version = client_python_version,
        dependency_settings = dependency_settings,
    )

def _handle_pre_exec_commands(commands: str):
    """Run the pre execution commands."""
    if commands == None or commands == "":
        logger.info("No pre-execution commands to run.")
    else:
        logger.info("Running pre-execution commands.")
        _run_shell_cmd(commands)
        logger.info("Finished running pre-execution commands.")


def _install_dependencies(
    dependency_file_dir: str,
    conda_env: str,
    client_python_version: str,
    dependency_settings: _DependencySettings = None,
):
    """Install dependencies in the job container

    Args:
        dependency_file_dir (str): Directory in the container where dependency file exists.
        conda_env (str): conda environment to be activated.
        client_python_version (str): Python version at the client side.
        dependency_settings (dict): Settings for installing dependencies.
    """

    if dependency_settings is not None and dependency_settings.dependency_file is None:
        # an empty dict is passed when no dependencies are specified
        logger.info("No dependencies to install.")
    elif dependency_settings is not None:
        dependencies_file = os.path.join(dependency_file_dir, dependency_settings.dependency_file)
        RuntimeEnvironmentManager().bootstrap(
            local_dependencies_file=dependencies_file,
            conda_env=conda_env,
            client_python_version=client_python_version,
        )
    '''
    else:
        # no dependency file name is passed when an older version of the SDK is used
        # we look for a file with .txt, .yml or .yaml extension in the workspace directory
        dependencies_file = None
        for file in os.listdir(dependency_file_dir):
            if file.endswith(".txt") or file.endswith(".yml") or file.endswith(".yaml"):
                dependencies_file = os.path.join(dependency_file_dir, file)
                break

        if dependencies_file:
            RuntimeEnvironmentManager().bootstrap(
                local_dependencies_file=dependencies_file,
                conda_env=conda_env,
                client_python_version=client_python_version,
            )
        else:
            logger.info(
                "Did not find any dependency file in the directory at '%s'."
                " Assuming no additional dependencies to install.",
                os.path.join(BASE_CHANNEL_PATH, channel_name),
            )
    '''


def _unpack_user_workspace():
    """Unzip the user workspace"""

    """
    workspace_archive_dir_path = os.path.join(BASE_CHANNEL_PATH, REMOTE_FUNCTION_WORKSPACE)

    if not os.path.exists(workspace_archive_dir_path):
        logger.info(
            "Directory '%s' does not exist.",
            workspace_archive_dir_path,
        )
        return None
    """

    workspace_archive_dir_path = runner_env.JOB_CWD

    workspace_archive_path = os.path.join(workspace_archive_dir_path, "workspace.zip")
    if not os.path.isfile(workspace_archive_path):
        logger.info(
            f"Workspace archive '{workspace_archive_dir_path}' does not exist.",
        )
        return None

    # workspace_unpack_dir = pathlib.Path(os.getcwd()).absolute()
    workspace_unpack_dir = runner_env.JOB_CWD
    

    print(f"_unpack_user_workspace(): workspace_archive_path: {workspace_archive_path}, workspace_unpack_dir: {workspace_unpack_dir}")
    '''
    shutil.unpack_archive(
        filename=workspace_archive_path, 
        extract_dir=workspace_unpack_dir,
        format = "zip"
    )
    '''
    _run_shell_cmd(f'unzip {workspace_archive_path} -d {workspace_unpack_dir}')
    logger.info(f"Successfully unpacked workspace archive at '{workspace_unpack_dir}'.")

    # workspace_unpack_dir = pathlib.Path(workspace_unpack_dir, JOB_REMOTE_FUNCTION_WORKSPACE)
    return workspace_unpack_dir


def _write_failure_reason_file(failure_msg):
    """Create a file 'failure' with failure reason written if bootstrap runtime env failed.

    See: https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html
    Args:
        failure_msg: The content of file to be written.
    """
    if not os.path.exists(FAILURE_REASON_PATH):
        with open(FAILURE_REASON_PATH, "w") as f:
            f.write("RuntimeEnvironmentError: " + failure_msg)




import typer
from cli.cmds.run import run
from cli.cmds.login import login
from cli.cmds.set_env import set_env
from cli.cmds.config import config
from cli.cmds.download import download
from cli.cmds.set_log_level import log_level
from cli.cmds.version import version
from cli.cmds.cmd_dump import dump
from cli.cmds.get import get
from cli.cmds.kill import kill
from cli.cmds.delete import delete
from cli.cmds.pd import pd
from cli.cmds.rec import rec

app = typer.Typer(help="A CLI tool to help you run your code in the Robbie")

app.command()(dump)
app.command()(run)
app.command()(login)
app.command()(set_env)
app.command()(config)
app.command()(download)
app.command()(version)
app.command()(get)
app.command()(kill)
app.command()(delete)
app.command()(pd)
app.command()(rec)
app.callback()(log_level)

if __name__ == "__main__":
    app()



import sys
import os
import subprocess
import platform
import psutil
from common.console import console, print_boxed_messages
from common.build_env import build_env
from common.utils import get_version
from common.env_defaults import current
from common.env_config import env
from common.user_config import user_config
from common.config import (
    load_job_config, 
    PositronJob
)
from common.cli_args import args
from common.api.validate_user_auth_token import is_auth_token_valid
from positron_job_runner.runtime_environment_manager import RuntimeEnvironmentManager, _DependencySettings, _python_executable

def common_dump(
    machine: bool = None,
    conda: bool = None,
    python: bool = None,
    sdk: bool = None,
    job_config: bool = None,
    job_config_name: str = "./job_config.yaml",
    save: bool = None
):

    def dump_machine():
        my_system = platform.uname()

        # Print the information
        my_info = f"System: {my_system.system}\n"
        my_info += f"Node Name: {my_system.node}\n"
        my_info += f"Release: {my_system.release}\n"
        my_info += f"Version: {my_system.version}\n"
        my_info += f"Machine: {my_system.machine}\n"
        my_info += f"Processor: {my_system.processor}\n"
        # Get CPU information
        my_info += f"CPU Count: {psutil.cpu_count(logical=False)}\n"
        my_info += f"CPU Usage: {psutil.cpu_percent(interval=1)}%\n"

        # Get memory information
        mem = psutil.virtual_memory()
        my_info += f"Total Memory: {mem.total / (1024.0 ** 3):.2f} GB\n"
        my_info += f"Available Memory: {mem.available / (1024.0 ** 3):.2f} GB\n"

        # Get disk information
        disk = psutil.disk_usage('/')
        my_info += f"Total Disk Space: {disk.total / (1024.0 ** 3):.2f} GB\n"
        my_info += f"Used Disk Space: {disk.used / (1024.0 ** 3):.2f} GB\n"

        print_boxed_messages("Local Machine Info", my_info)

    def dump_conda():
        console.print("Dumping Conda environment")

        is_conda_installed = False
        try:
            RuntimeEnvironmentManager()._get_conda_exe()
            is_conda_installed = True
        except ValueError as e:
           console.print(f"[red]No Conda/mamba executable not found: {str(e)}")

        # sanity checks
        if not is_conda_installed and os.getenv("CONDA_DEFAULT_ENV"):
            console.print(f"No Conda executable found but you are running in a Conda environment({RuntimeEnvironmentManager._get_active_conda_env_name()})")
            return
    
        if is_conda_installed and not os.getenv("CONDA_DEFAULT_ENV"):
            console.print(f"Conda executable found but you are not running in a conda environment (it's OK)")
            return

        if is_conda_installed and os.getenv("CONDA_DEFAULT_ENV"):
            result = subprocess.run("conda info", capture_output=True, shell=True, text=True)
            print_boxed_messages(f'conda info', result.stdout)

            title = f"Conda environment detected: {RuntimeEnvironmentManager()._get_active_conda_env_name()}"
            py_env = f'CONDA_DEFAULT_ENV (active): {RuntimeEnvironmentManager()._get_active_conda_env_name()}\n'
            py_env += f'CONDA_PREFIX: {RuntimeEnvironmentManager()._get_active_conda_env_prefix()}\n'
            py_env += f'conda_executable: {RuntimeEnvironmentManager()._get_conda_exe()}\n'
            try:
                py_conda = RuntimeEnvironmentManager()._python_version_in_conda_env(os.getenv("CONDA_DEFAULT_ENV"))
                py_env += f'python_version_in_conda_env: {py_conda}\n'
            except Exception as e:
                py_env += f'python_version_in_conda_env: Python not found\n'

            py_env = f'Python executable (sys.executable): {sys.executable}\n'
            py_env += f'Python executable (os.__file__): {os.__file__}\n'
            py_env += f'current_python_version: {RuntimeEnvironmentManager()._current_python_version()}\n'

            print_boxed_messages(f'({title})', py_env)

            result = subprocess.run("conda list", capture_output=True, shell=True, text=True)
            print_boxed_messages(f'Installed Python Packages (% conda list)', result.stdout)

            result = subprocess.run("conda config --show", capture_output=True, shell=True, text=True)
            print_boxed_messages(f'Conda configuration (% conda config --show)', result.stdout)

            # result = subprocess.run("conda env export", capture_output=True, shell=True, text=True)
            # print_boxed_messages("% conda env export", result.stdout)

            result = subprocess.run("conda env export --from-history", capture_output=True, shell=True, text=True)
            print_boxed_messages("% conda env export --from-history", result.stdout)

            # Try to capture dependencies from the conda environment, if any.
            conda_env_prefix = RuntimeEnvironmentManager()._get_active_conda_env_prefix()
            local_dependencies_path = os.path.join(os.getcwd(), "dumped_environment.yml")
            RuntimeEnvironmentManager()._export_conda_env_from_prefix(conda_env_prefix, local_dependencies_path)
            
            try:
                with open(local_dependencies_path, 'r') as file:
                    # Read the content of the file
                    file_content = file.read()
                    print_boxed_messages(f"Platform-specific: ({platform.system()}) export: {local_dependencies_path}", file_content)
            except Exception as e:
                console.print(f'[red]Failed to open: {local_dependencies_path}')

    def dump_python():
        console.print("Dumping Python environment")
        py_env = f'Python executable (sys.executable): {sys.executable}\n'
        py_env += f'Python executable (os.__file__): {os.__file__}\n'
        py_env += f'current_python_version: {RuntimeEnvironmentManager()._current_python_version()}\n'
    
        if not os.getenv("CONDA_DEFAULT_ENV"):
            print_boxed_messages("Non-Conda Conda environment detected", py_env)
            if platform.system() == 'Windows':
                result = subprocess.run("where python", capture_output=True, shell=True, text=True)
                print_boxed_messages('Python local (% where python)', result.stdout)
            else:   
                result = subprocess.run("which python", capture_output=True, shell=True, text=True)
                print_boxed_messages('Python local (% which python)', result.stdout)
            result = subprocess.run("pip list", capture_output=True, shell=True, text=True)
            print_boxed_messages('Installed Python Packages (% pip list)', result.stdout)

            # show the auto-capture output into requirements.txt
            RuntimeEnvironmentManager()._capture_from_pip_runtime(output_file="dump_requirements.txt")
            try:
                with open("dump_requirements.txt", 'r') as file:
                    # Read the content of the file
                    file_content = file.read()
                    print_boxed_messages(f"Auto-captured dump_requirements.txt", file_content)
            except Exception as e:
                console.print(f"[red]Failed to open: 'dump_requirements.txt'")
        else:
            console.print("[red]Conda environment detected, skipping Python local details. Run `robbie dump --conda` to see Conda details")

    def dump_sdk():
        console.print("Dumping SDK environment")
        # now dump everything related to the Robbie SDK version
        rob_ver = f'current_robbie_pysdk_version: {RuntimeEnvironmentManager()._current_robbie_pysdk_version()}\n'
        result = subprocess.run("pip show robbie", capture_output=True, shell=True, text=True)
        rob_ver += f"Robbie SDK version (% pip show robbie): {result.stdout}"
        rob_ver += f"Robbie Python SDK version - importlib.metadata.version('robbie'): {get_version()}"
        print_boxed_messages("Robbie Version Details", rob_ver)

        # dump the SDK environment
        print_boxed_messages("build_env", str(build_env))
        print_boxed_messages("EnvConfig (env)", env.to_string())
        print_boxed_messages("EnvDefaults (current)", current.to_string())
        print_boxed_messages("User Config (user_config)", user_config.to_string())     
        print_boxed_messages("CLI Arguments (args)", args.to_string())

        print_boxed_messages("API Test", api_test())

    def dump_job_config(
        dump: bool,
        name: str = None
    ):

        if dump == False:
            return

        if name == None:
            name = "./job_config.yaml"

        console.print(f"Dumping Job Config: {name}")
        try:
            global_job_config_yaml = load_job_config(name)
        except Exception as e:
            print(f"Error loading job configuration: {name} {str(e)}")
            global_job_config_yaml = None
    
        if global_job_config_yaml:
            print_boxed_messages(f"{name} contents:", global_job_config_yaml.to_string())

            if global_job_config_yaml.include_local_dir:
                contents = os.listdir(os.getcwd())
                print_boxed_messages('CWD (os.getcwd()) contents', "\n".join(contents))

            if global_job_config_yaml.dependencies and global_job_config_yaml.dependencies != "none":

                # return the file name or auto-capture
                try:
                    dep_file = RuntimeEnvironmentManager().snapshot(global_job_config_yaml.dependencies)

                    if os.path.exists(dep_file):
                        try:
                            with open(dep_file, 'r') as file:
                                # Read the content of the file
                                file_content = file.read()
                                print_boxed_messages(f'global_job_config_yaml.dependencies: {global_job_config_yaml.dependencies}, file: {dep_file}', file_content)
                        except Exception as e:
                            console.print(f'[red]Failed to open: global_job_config_yaml.dependencies: {dep_file}')
                    else:
                        console.print(f'[red]global_job_config_yaml.dependencies file: {dep_file} does not exist')
                except Exception as e:
                    console.print(f'[red]Error capturing dependencies: {str(e)}')
        

    if machine:
        dump_machine()
    if conda:
        dump_conda()
    if python:
        dump_python()
    if sdk:
        dump_sdk()
    if job_config:
        dump_job_config(job_config, job_config_name)

    # print(f"machine: {machine}, conda: {conda}, python: {python}, sdk: {sdk}, job_config: {job_config}, job_config_name: {job_config_name}")

    flag = machine or conda or python or sdk or job_config
    if not flag and job_config_name:
        console.print(f"[red]Error: job_config_name: {job_config_name} provided but no other flags set. Did you mean --conda or --python?")
    elif not flag:
        dump_machine()
        if os.getenv("CONDA_DEFAULT_ENV"):
            dump_conda()
        else:
            dump_python()
        dump_sdk()
        dump_job_config(job_config, job_config_name)

    print_boxed_messages("End of Robbie local environment dump", "That's all folks!")


def dump_data():
    """All the same data as common_dump, but returned as a dict"""
    fake_job_config = False
    try:
        job_config = load_job_config()
        if not job_config:
            fake_job_config = True
            job_config = PositronJob()
            job_config.dependencies = None
    except Exception as e:
        print(f'Error loading job configuration! {str(e)}')
        fake_job_config = False

    return {
        "Python executable (sys.executable)": sys.executable,
        "Python executable (os.__file__)": os.__file__,
        "Robbie Python SDK version - importlib.metadata.version('robbie')": get_version(),
        "Robbie SDK version (% pip show robbie)": subprocess.run("pip show robbie", capture_output=True, shell=True, text=True).stdout,
        "Installed Python Packages (% pip list)": subprocess.run("pip list", capture_output=True, shell=True, text=True).stdout,
        "./requirements.txt": open("./requirements.txt", 'r').read() if os.path.exists("./requirements.txt") else None,
        "./job_config.yaml": job_config.to_string() if not fake_job_config else "No job_config.yaml found",
        # "RuntimeEnvironmentManager": rtem_to_string(job_config),
        "build_env": str(build_env),
        "EnvConfig (env)": env.to_string(),
        "EnvDefaults (current)": current.to_string(),
        "User Config (user_config)": user_config.to_string(),
        "CLI Arguments (args)": args.to_string(),
    }

def api_test():
    try:
        result = is_auth_token_valid()
        return f"is_auth_token_valid(): API call - successful, result: {result}"
    except Exception as e:
        return f"is_auth_token_valid(): API call - failed: {str(e)}"




import os
import argparse
import pyfiglet
import platform
from functools import wraps
from common.cli_args import args as cli_args
from common.config import (
    PositronJob, 
    merge_from_yaml_and_args,
    load_job_config
)
from remote_function.deploy import Deploy
from remote_function.stored_function import StoredFunction
from common.observability.main import track_command_usage
from common.image_name import get_auto_image_name_and_cluster
from common.user_config import user_config
from common.console import console, ROBBIE_BLUE
from common.logging_config import logger, set_log_level
from common.enums import JobRunType
from common.utils import _exit_by_mode, _nb
from common.env_defaults import current
from robbie.notebook_cell_ui import notebook_cell_ui
from cli.cmds.login import login
from positron_job_runner.runtime_environment_manager import (
    RuntimeEnvironmentManager, 
    _build_job_args, 
    _running_in_conda,
    CONDA_MODE,
    PYTHON_MODE,
    GENERIC_MODE
)

def remote(**parameters):
    """
    Decorator to deploy a function in Robbie. 
    This works in both a Jupyter notebook and a Python script.

    You can pass argumments to the decorator to customize the deployment in two ways:
    - Command Line Arguments when running from the commmand line - e.g. python my_script.py --tail
        Supported arguments:
            --tail: bool
            --loglevel: str [CRITICAL,FATAL,ERROR, WARNING, INFO, DEBUG, NOTSET]
            --create-only: bool
            - local_container: str (path to your distribution)
            --results-from-job-id: str (job_id )
    - Decorator Aguments in a Python script or Jupyter Notebook when decorating a function - e.g. @remote(tail=True)
        Supported arguments:
            - funding_group_id: str
            - environment_id: str
            - image: str
            - tail: bool    
            - loglevel: str
            - create_only: bool
            - chooser_ui: bool (Enable user to choose job config via a small GUI in the cell - Used in Jupyter Notebook only)

    """
    if not os.getenv('POSITRON_CLOUD_ENVIRONMENT'):
        # Parse command line arguments
        parser = argparse.ArgumentParser(description = "A decorator to handle deploying running your function in the cloud")
        parser.add_argument('--f', type=str, help='Path to job_config.yaml file', dest='job_config_arg', default='./job_config.yaml')
        parser.add_argument('--tail', action='store_true', help='Stream the stdout from Robbie back to your cli', dest='stream_stdout', default=False)
        parser.add_argument('--loglevel', help='Set the logging level [CRITICAL,FATAL,ERROR, WARNING, INFO, DEBUG, NOTSET]')
        parser.add_argument('--create-only', action='store_true', help='Create the job but do not run it.', dest='create_only')
        parser.add_argument('--local', type=str, help='Run job on local container <path to distribution>', dest='local_container')
        parser.add_argument('--results-from-job-id', help='Fetch results and return from decorated function.')
        parser.add_argument('--v', action='store_true', help='Verbose logging.', dest='verbose')
        positron_args, _ = parser.parse_known_args()

        verbose_arg = False

        if positron_args.loglevel:
            set_log_level(positron_args.loglevel)
        if positron_args.verbose:
            verbose_arg = True

        ascii_banner = pyfiglet.figlet_format("Robbie")
        console.print(ascii_banner, style='#41a7ff')

        console.print(f"[blue]Platform: [white]{platform.system()}[blue] Stage: [white]{current.name}")
        if platform.system() == "Windows" and os.getenv('CONDA_DEFAULT_ENV'):
            console.print(f"[yellow] Windows conda support is currently experimental.")

        logger.debug("========positron_args========")
        # print("positron_args=", positron_args)
        logger.debug(positron_args) 

        # Jupyter Support - Default out the cli_args to run remote always with no prompting
        if not cli_args.is_init:
            cli_args.init(
                local=False,
                deploy=True,
                stream_stdout=positron_args.stream_stdout,
                create_only=positron_args.create_only,
                local_container=positron_args.local_container,
                results_from_job_id=positron_args.results_from_job_id,
                skip_prompts=True,
            )

        chooser_ui_flag = False
        # enable  and tail function parameters but remove them before passing to PositronJob config
        if "loglevel" in parameters:
            set_log_level(parameters["loglevel"])
            del parameters["loglevel"]
        if "tail" in parameters:
            cli_args.stream_stdout = parameters["tail"]
            del parameters["tail"]
        if "create_only" in parameters:
            cli_args.create_only = parameters["create_only"]
            del parameters["create_only"]
        if "results_from_job_id" in parameters:
            cli_args.results_from_job_id = parameters["results_from_job_id"]
            del parameters["results_from_job_id"]
        if "chooser_ui" in parameters:
            chooser_ui_flag = True
            del parameters["chooser_ui"]
        if "verbose" in parameters:
            verbose_arg = True
            del parameters["verbose"]

        console.print("[green]✔[/green] Current working directory: ", os.getcwd(), style=ROBBIE_BLUE)
        
        # was there a job_config.yaml file in the CWD?
        try: 
            global_job_config_yaml = load_job_config(positron_args.job_config_arg)
            if global_job_config_yaml:
                logger.debug(f"global_job_config_yaml: {global_job_config_yaml.to_string()}")
                console.print(f"[green]✔[/green] Loaded run configuration file: ", end="", style=ROBBIE_BLUE)
                console.print(f"{positron_args.job_config_arg}")
                if global_job_config_yaml.job_type != JobRunType.REMOTE_FUNCTION_CALL:
                    console.print(f"[red]Error, job_config.yaml file is not configured for a remote function.")
                    return
            else:
                console.print("[yellow] No job_config.yaml file found.")
                logger.debug("No job_config.yaml file found.")
        except Exception as e:
            print(f'Error loading job configuration! {str(e)}')
            return

        # print(f"global_job_config_yaml: {global_job_config_yaml.to_string()}")

        if _running_in_conda():
            console.print(f"[green]✔[/green] Running in conda environment: {os.getenv('CONDA_DEFAULT_ENV')}", style=ROBBIE_BLUE)
        else:
            console.print(f"[green]✔[/green] Running in non-conda environment", style=ROBBIE_BLUE)


    def decorator(func):
        @track_command_usage("remote")
        @wraps(func)
        def wrapper(*args, **kwargs):
            logger.debug("Running decorator")

            if os.getenv('POSITRON_CLOUD_ENVIRONMENT'):
                logger.debug("Running function locally")
                return func(*args, **kwargs)

            # This check eliminates a extra step if the user happens to run robbie run and the API key is not valid
            login()

            if cli_args.results_from_job_id:
                stored_function = StoredFunction(func, args, kwargs)
                stored_function.set_job_id(cli_args.results_from_job_id)
                secret_key = user_config.user_auth_token if user_config.user_auth_token else ""
                stored_function.load_and_validate_results(hmac_key=secret_key)
                return stored_function.result

            # get decorator arguments
            job_config = None
            job_config_ui_or_arguments = None
            nonlocal chooser_ui_flag
            
            if chooser_ui_flag:
                if _nb: # we are in a notebook
                    job_config_ui_or_arguments = notebook_cell_ui()
                    if job_config_ui_or_arguments == None:
                        console.print("[red] User interrupted.")
                        return
                else:
                    console.print("[yellow]Warning: The 'chooser_ui' is only supported in Jupyter Notebooks. Please remove this argument.[/yellow]")
                    logger.warning("Warning: The 'chooser_ui' is only supported in Jupyter Notebooks. Please remove this argument.")
                    _exit_by_mode(1)
                logger.debug(job_config_ui_or_arguments.to_string("job_config_ui_or_arguments (arguments passed into remote function)"))
            else:
                job_config_ui_or_arguments = PositronJob(**parameters)
                if job_config_ui_or_arguments == None:
                    console.print("[red]Failed to create PositronJob from @remote parameters.")
                    return
                
                # print(f"job_config_ui_or_arguments: {job_config_ui_or_arguments.to_string()}")
                
                if job_config_ui_or_arguments.rpv:
                    console.print("[red]Error: The 'python_version' argument is not supported with remote functions.\nPlease remove it.[/red]")
                    _exit_by_mode(1)

                # track where the parameters come from so we can display to the user later
                if job_config_ui_or_arguments.name:
                    # console.print(f"name: {job_config_ui_or_arguments.name} - argument to @remote decorator.")
                    job_config_ui_or_arguments.name_selection = "argument to @remote decorator"
                if job_config_ui_or_arguments.funding_group_id:
                    # console.print(f"funding_group_id: {job_config_ui_or_arguments.funding_group_id} - argument to @remote decorator.")
                    job_config_ui_or_arguments.funding_selection = "argument to @remote decorator"
                if job_config_ui_or_arguments.environment_id:
                    # console.print(f"environment_id: {job_config_ui_or_arguments.environment_id} - argument to @remote decorator.")
                    job_config_ui_or_arguments.environment_selection = "argument to @remote decorator"
                if job_config_ui_or_arguments.image:
                    # console.print(f"image: {job_config_ui_or_arguments.image} - argument to @remote decorator.")
                    job_config_ui_or_arguments.image_selection = "argument to @remote decorator"
                if job_config_ui_or_arguments.dependencies:
                    # console.print(f"deps: {job_config_ui_or_arguments.dependencies} - argument to @remote decorator.")
                    job_config_ui_or_arguments.dep_selection = "argument to @remote decorator"
                if job_config_ui_or_arguments.include_local_dir:
                    # console.print(f"include_local_dir: {job_config_ui_or_arguments.include_local_dir} - argument to @remote decorator.")
                    job_config_ui_or_arguments.include_local_dir_selection = "argument to @remote decorator"


                logger.debug(job_config_ui_or_arguments.to_string("job_config_ui_or_arguments (arguments passed into remote function)"))


            if global_job_config_yaml and global_job_config_yaml.rpv:
                console.print("[red]Error: You cannot set 'rpv' in job_config.yaml. Please remove it.[/red]")

            job_config = merge_from_yaml_and_args(global_job_config_yaml, job_config_ui_or_arguments)

            if not job_config:
                console.print('[red]Error: Unable to merge yaml and arguments/ui selection.')
                return

            if job_config.commands:
                console.print("[red]Error: The 'commands' configuration in job_config.yaml is not supported in the remote decorator.\nPlease remove it or run with 'robbie run' to use 'commands'.[/red]")
                logger.error("The 'commands' configuration in job_config.yaml is not supported in the remote decorator.")
                _exit_by_mode(1)

            if job_config.rpv and job_config.rpv != "local":
                console.print("[yellow]Warning: You cannot set the remote python in the job_config.yaml or argument when running a remote function...ignoring.\n")
                logger.error("The 'python_version' configuration in job_config.yaml is not supported in the remote decorator.")
                _exit_by_mode(1)
        
            job_config.job_type = JobRunType.REMOTE_FUNCTION_CALL
            job_config.rpv = RuntimeEnvironmentManager()._current_python_version()
            job_config.rpv_selection = "Auto-detected from local machine"
            logger.debug(job_config.to_string("job_config being passed to function"))

            # run with no job_config.yaml or agument
            if not job_config.mode:
                if _running_in_conda():
                    job_config.mode = CONDA_MODE
                    job_config.mode_selection = "not set, auto-detected"
                else:
                    job_config.mode = PYTHON_MODE
                    job_config.mode_selection = "not set, auto-detected"
                
            # if deps are not set, auto-capture
            if not job_config.dependencies:
                console.print(f"No deps set, auto-capturing deps", style=ROBBIE_BLUE)
                job_config.dependencies = "auto-capture"

            if job_config.dependencies == "auto-capture":
                job_config.dep_selection = "auto-captured"

            # sanity checking
            if _running_in_conda() and job_config.mode == PYTHON_MODE:
                console.print("[red]Error: You can't run a non-conda Python job in a conda environment.[/red]")
                _exit_by_mode(1)

            # sanity checking
            if not _running_in_conda() and job_config.mode == CONDA_MODE:
                # we have a problem if we are in a non-conda environment
                console.print("[red]Error: You can't specify conda deps in non-conda environment. Please use requirement.txt[/red]")
                _exit_by_mode(1)

            console.print("Robbie is analyzing run information...", style=ROBBIE_BLUE)

            # snapshot_freeze() does all of the dependency capturing
            job_config.dependencies = RuntimeEnvironmentManager().snapshot(job_config.dependencies)

            if _running_in_conda() and job_config.dependencies.endswith(".txt"):
                console.print("[red]Error: You can't use a requirements.txt file in a conda environment. Please use a .yaml file instead.[/red]")
                _exit_by_mode(1)
            
            if not _running_in_conda() and (job_config.dependencies.endswith(".yml") or job_config.dependencies.endswith(".yaml")):
                console.print("[red]Error: You can't specify conda deps in non-conda environment. Please use requirement.txt[/red]")
                _exit_by_mode(1)
            console.print("[green]✔[/green] Determined environment", style=ROBBIE_BLUE)

            #
            # Image selection
            #
            if job_config.image == "auto-select":
                job_config.image, job_config.cluster = get_auto_image_name_and_cluster(
                    job_config.funding_group_id, 
                    job_config.environment_id,
                    job_config.rpv
                )
                job_config.image_selection = "auto-selected"
                # console.print('Auto-selecting image:', style=ROBBIE_BLUE, end="")
                # console.print(f'{jctr.image} on {jctr.cluster} cluster', style=ROBBIE_DEFAULT)
                console.print("[green]✔[/green] Automatically selected image", style=ROBBIE_BLUE)


            # note: we have changed the behavior or job_args
            cli_args.job_args = _build_job_args(
                mode = job_config.mode,
                rpv = job_config.rpv, 
                dependencies = job_config.dependencies
            )
            logger.debug(f"job_args: {cli_args.job_args}")

            # verbose printouts
            if verbose_arg:
                job_config.verbose = True

            console.print(f"Robbie is running your function remotely!", style="bold")
            return Deploy.remote_function_deploy(func, args, kwargs, job_config)

        return wrapper
    return decorator


from pydantic import BaseModel, SerializeAsAny, field_validator
from typing import Dict, Optional, List, Union
import os
import copy
import yaml
import re
from yaml import dump
from enum import Enum
from common.utils import get_version, _valid_python_versions
from common.exceptions import RobbieException
from common.user_config import user_config
from common.constants import JOB_CONF_YAML_PATH, RUNTIME_ENV_PREFIX
from common.logging_config import logger
from common.enums import JobRunType
from positron_job_runner.runtime_environment_manager import (
    CONDA_MODE,
    PYTHON_MODE,
    GENERIC_MODE,
    _running_in_conda
)
from common.console import console
from common.image_name import get_auto_image_name_and_cluster
from common.api.funding_envs_images import *

backward_comptability_1_0 = False

class PositronJob(BaseModel):
    """
    The Job details as defined in the `python_job` from the `job_config.yaml` file.
    """
    name: Optional[str] = None
    job_type: Optional[JobRunType] = None
    mode: Optional[str] = None
    rpv: Optional[str] = None # only set if we change the python version
    funding_group_id: Optional[str] = None
    environment_id: Optional[str] = None
    cluster: Optional[str] = None
    image: Optional[str] = None # this is not written to the config file
    workspace_dir: Optional[str] = None # deprecated
    include_local_dir: Optional[bool] = False
    custom_file_filter: Optional[List[str]] = None
    dependencies: Optional[str] = None
    max_tokens: Optional[int] = None
    max_time: Optional[int] = None
    env: Optional[Dict[str, str]] = None
    commands: Optional[List[str]] = None
    # below are values not written to the config file
    filename: Optional[str] = None
    robbie_sdk_version: Optional[str] = get_version()
    image_selection: Optional[str] = None
    funding_selection: Optional[str] = None
    environment_selection: Optional[str] = None
    rpv_selection: Optional[str] = None
    dep_selection: Optional[str] = None
    name_selection: Optional[str] = None
    include_local_dir_selection: Optional[str] = None
    mode_selection: Optional[str] = None
    verbose: Optional[bool] = False
        
    # remote python version
    @field_validator('rpv', mode='after')
    def ensure_valid_python_version(cls, rpv):
        rpv_list = _valid_python_versions()
        rpv_list.append("local")
        if rpv in rpv_list:
            return rpv
        else:
            raise ValueError('Invalid Python version, must be "3.9", "3.10, "3.11", or "3.12"')
        
    @field_validator('dependencies', mode='after')
    def ensure_valid_dependies(cls, dependencies):
        if (dependencies == "auto-capture" or
            dependencies == "none" or 
            dependencies.endswith(".txt") or 
            dependencies.endswith(".yml") or 
            dependencies.endswith(".yaml")):
            return dependencies
        else:
            raise ValueError('Invalid dependencies. Must be "auto-capture", "none", or end with ".txt", ".yml", or ".yaml"')
        
    @field_validator('commands', mode='after')
    def ensure_non_empty(cls, commands):
        return commands if len(commands) else None

    @field_validator('env', mode='after')
    def ensure_env_is_dict(cls, v):
        if isinstance(v, dict):
            return v
        raise ValueError('env must be a dictionary')

    @field_validator('workspace_dir', mode='after')
    def workspace_dir_deprecation_warning(cls, workspace_dir):
        console.print("[yellow] Warning: workspace_dir is deprecated. Please use include_local_dir instead.")
        return workspace_dir
    
    @field_validator('max_time', mode='before')
    def ensure_max_time_is_int(cls, max_time: Union[int, str, None]) -> Union[int, None]:
        return cls._max_time_to_minutes(max_time)

    @field_validator('max_tokens', mode='before')
    def ensure_max_tokens_is_int(cls, max_tokens: Union[int, str, None]) -> Union[int, None]:
        return cls._max_tokens_to_int(max_tokens)

    def create_runtime_env(self) -> Dict[str, str]:
        """
        Used on the client side to create the prefixed runtime environment variables
        to avoid conflicts with the local environment variables.
        """
        env: Dict[str, str] = {}
        # cycle through the env vars and prefix them
        if not self.env:
            return env
        for key, value in self.env.items():
            if (value == ""):
                env_var = os.environ.get(key)
                if env_var is None:
                    raise ValueError(f"The env prop {key} is unset inside job_config.yaml and also unset in local env vars. Please set this value.")
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = env_var
            else:
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = value
        return env

    @staticmethod
    def _max_time_to_minutes(max_time: Union[int, str, None]) -> Union[int, None]:
        if not max_time:
            return None
        if isinstance(max_time, int):
            return max_time
        matches = re.search(r'^(\d+):(\d{2})$', max_time)
        if matches is None:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        try:
            hours = int(matches.group(1))
            minutes = int(matches.group(2))
        except:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        if minutes >= 60:
            raise ValueError('Invalid Job Config: Field "max_time" ({max_time}) has invalid minutes! Must be 0 <= minutes < 60!')
        return hours * 60 + minutes

    @staticmethod
    def _max_tokens_to_int(max_tokens: Union[int, str, None]) -> Union[int, None]:
        if not max_tokens:
            return None
        if isinstance(max_tokens, int):
            return max_tokens
        try:
            max_tokens = int(max_tokens)
        except:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        if max_tokens <= 0:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        return max_tokens

    def validate_values(self) -> None:
        errors = []
        if self.env and not validate_env_vars(self.env):
            errors.append('At least one of the environment variables provided is invalid')
        if errors:
            raise RobbieException(f'Invalid configuration. Errors: {errors}')
        return None
    
    def namestr(self, obj, namespace):
        return [name for name in namespace if namespace[name] is obj]

    def to_string(self, title: str = None):
        message = f"""
- filename: {self.filename}
- name: {self.name}
    - name_selection: {self.name_selection}
- job_type: {self.job_type}
- mode: {self.mode}
    - mode_selection: {self.mode_selection}
- rpv: {self.rpv}
    - rpv_selection: {self.rpv_selection}
- robbie_sdk_version: {self.robbie_sdk_version}
- funding_group_id: {self.funding_group_id}
    - funding_selection: {self.funding_selection}
- environment_id: {self.environment_id}
    - environment_selection: {self.environment_selection}
- image: {self.image}
    - image_selection: {self.image_selection}
- cluster: {self.cluster}
- dependencies: {self.dependencies}
    - dep_selection: {self.dep_selection}
- custome_file_filter: {self.custom_file_filter}
- verbose: {self.verbose}
- workspace_dir: {self.workspace_dir}
- include_local_dir: {self.include_local_dir}
    - include_local_dir_selection: {self.include_local_dir_selection}
- max_tokens: {self.max_tokens}
- max_time: {self.max_time}
- env: {self.env}
- commands: {self.commands}"""

        if title:
            return f"========== {title} ==========\n{message}"
        else:
            return message
        
    def deepcopy(self):
        return copy.deepcopy(self)
        

class PositronJobConfig(BaseModel):
    """
    The `job_config.yaml` schema class.
    """
    version: float
    python_job: PositronJob

    def write_to_file(this, filename: str = JOB_CONF_YAML_PATH):
        copy_of_config = copy.deepcopy(this)
        if copy_of_config.python_job.rpv == "local":
            del copy_of_config.python_job.rpv
        del copy_of_config.python_job.robbie_sdk_version
        del copy_of_config.python_job.cluster
        del copy_of_config.python_job.image_selection
        del copy_of_config.python_job.funding_selection
        del copy_of_config.python_job.environment_selection
        del copy_of_config.python_job.rpv_selection
        del copy_of_config.python_job.dep_selection
        del copy_of_config.python_job.name_selection
        del copy_of_config.python_job.include_local_dir_selection
        del copy_of_config.python_job.mode_selection
        del copy_of_config.python_job.verbose
        del copy_of_config.python_job.filename

        config_dict = copy_of_config.model_dump(
            exclude_unset=True,
        )
        config_dict = convert_enums_to_values(config_dict)

        with open(filename, 'w') as file:
            file.write(dump(config_dict, sort_keys=False))


def convert_enums_to_values(d: dict) -> dict:
    """
    Converts Enum type values in the dictionary to their respective values.
    """
    for key, value in d.items():
        if isinstance(value, Enum):
            d[key] = value.value
        elif isinstance(value, dict):
            convert_enums_to_values(value)
    return d

def convert_enums_to_values(d: dict) -> dict:
    """
    Converts Enum type values in the dictionary to their respective values.
    """
    for key, value in d.items():
        if isinstance(value, Enum):
            d[key] = value.value
        elif isinstance(value, dict):
            convert_enums_to_values(value)
    return d

def is_valid_key_value(keyvalue):
    """
    Validate that the key-value contains only alphanumeric characters, dashes, and underscores, and has no spaces.
    """
    return bool(re.match(r'^[\w-]+$', keyvalue))

def validate_env_vars(env_dict):
    """
    Validate the environment variables from the given dictionary.
    """
    valid = True
    for key, value in env_dict.items():
        if not is_valid_key_value(key):
            print(f"Invalid key (contains invalid characters or spaces): {key}")
            valid = False
        if value != "" and not is_valid_key_value(value):
            print(f"Invalid value (contains invalid characters or spaces): {value}")
            valid = False
    return valid

def merge_config(base_config: PositronJob, override_config: PositronJob) -> PositronJob:
    """
    Makes it easy to merge decorator configs on top of the YAML config.
    """
    update_data = override_config.model_dump(exclude_unset=True)
    updated_config = base_config.model_copy(update=update_data)
    return updated_config


def invalid_yaml_keys(yaml_job_keys) -> bool:
    """
    Validates the yaml keys against the PositronJob class.
    """
    # Get only attributes from PositronJob
    base_attrs = set(dir(BaseModel))
    derived_attrs = set(dir(PositronJob()))
    additional_attrs = derived_attrs - base_attrs

    # Exclude built-in attributes (e.g., __init__, __module__)
    validKeys = [attr for attr in additional_attrs if not attr.startswith('__')]
    
    weHaveInvalidYamlKeys = False
    for key in yaml_job_keys:
        if key not in validKeys:
            weHaveInvalidYamlKeys = True
            raise RobbieException(f'Error: wrong param in the job_config.yaml: -> {key}')

    return weHaveInvalidYamlKeys

def load_job_config(config_path: str = './job_config.yaml') -> Optional[PositronJob]:
    """
    Load the job configuration from the `job_config.yaml` file if it exists
    """
    if not os.path.exists(config_path):
        logger.debug(f'{config_path} file not found')
        return None
    
    try:
        # this happens sometimes and it causes the job to fail
        if os.path.getsize(config_path) == 0:
            logger.debug(f'{config_path} file is empty')
            return None
        with open(config_path, 'r') as job_config_file:
            job_config_dict = yaml.safe_load(job_config_file)
            job = job_config_dict["python_job"]
            if invalid_yaml_keys(job.keys()):
                return None
            job_config = PositronJobConfig(**job_config_dict)

            logger.debug(f"job_config.version: {job_config.version}")
            if job_config:
                job_config.python_job.filename = config_path
                if job_config.version == 1.0:
                    global backward_comptability_1_0
                    backward_comptability_1_0 = True

                    console.print(f"[bold yellow]Warning: Your job_config.yaml file is using an older version {job_config.version} than current 1.1.")
                    console.print("[bold yellow]To get the most functinality, please run `robbie configure` or `robbie run --i` to create the lastest job_config.yaml file.")
                    console.print("[bold yellow]Robbie will attempt run your job with defaults, but some features may not work.")

                    if job_config.python_job.commands:
                        console.print(f"[yellow]job_config.yaml contains the following commands:")
                        for cmd in job_config.python_job.commands:
                            console.print(f"[yellow] {cmd}")
                        console.print(f"[yellow]Defaulting to JobType: BASH_COMMAND_RUNNER")
                        job_config.python_job.job_type = JobRunType.BASH_COMMAND_RUNNER
                        job_config.python_job.mode = GENERIC_MODE
                        if _running_in_conda():
                            console.print("[yellow]Detected you are running in a conda environment...ignoring.")
                        else:
                            console.print("[yellow]Not running in a conda environment...ignoring")
                    else:
                        console.print("[yellow]  - No commands found, defaulting to REMOTE_FUNCTION_CALL")
                        job_config.python_job.job_type = JobRunType.REMOTE_FUNCTION_CALL
                        if _running_in_conda():
                            console.print(f"[yellow]Detected you are running in a conda environment {os.environ['CONDA_PREFIX']}")
                            job_config.python_job.mode = CONDA_MODE
                        else:
                            console.print("[yellow]Not running in a conda environment...ignoring")
                            job_config.python_job.mode = PYTHON_MODE
                    
                    if job_config.python_job.workspace_dir:
                        console.print("[yellow]  - workspace_dir is deprecated. Setting `include_local_dir` = True")
                        job_config.python_job.include_local_dir = True

                elif job_config and job_config.version == 1.1:
                    if job_config.python_job.workspace_dir:
                        console.print("[yellow]  - workspace_dir is deprecated. Please use include_local_dir instead.")
                        job_config.include_local_dir = True

                if job_config.python_job.name:
                    job_config.python_job.name_selection = f"from {config_path}"
                if job_config.python_job.funding_group_id:
                    job_config.python_job.funding_selection = f"from {config_path}"
                if job_config.python_job.environment_id:
                    job_config.python_job.environment_selection = f"from {config_path}"
                if job_config.python_job.image:
                    job_config.python_job.image_selection = f"from {config_path}"
                if job_config.python_job.dependencies:
                    job_config.python_job.dep_selection = f"from {config_path}"
                if job_config.python_job.rpv:
                    job_config.python_job.rpv_selection = f"from {config_path}"
                if job_config.python_job.include_local_dir:
                    job_config.python_job.include_local_dir_selection = f"from {config_path}"
                if job_config.python_job.mode:
                    job_config.python_job.mode_selection = f"from {config_path}"

                return job_config.python_job
            else:
                return None

    except Exception as e:
        raise Exception(e)



def merge_from_yaml_and_args(
    input_job_config: Union[PositronJob, None],
    args_job_config: Union[PositronJob, None]
) -> PositronJob:
    """
    Merge the job_config (from the yaml or empty object) file with the command line arguments for funding, environment, and image.
    Ensure that environment is in the funding group.

    Behavior:
    - Command line arguments take precedence over the job_config.yaml file
    - Command line arguments are not memorized in the job_config.yaml file 
    - Defaults are applied where selections are missing

    """
    # if arges_job_config is None, then we are not using the command line arguments
    name_arg = args_job_config.name if args_job_config else None
    funding_arg = args_job_config.funding_group_id if args_job_config else None
    environment_arg = args_job_config.environment_id if args_job_config else None
    image_arg = args_job_config.image if args_job_config else None
    include_local_dir_arg = args_job_config.include_local_dir if args_job_config else None
    dependencies_arg = args_job_config.dependencies if args_job_config else None
    mode_arg = args_job_config.mode if args_job_config else None
    rpv_arg = args_job_config.rpv if args_job_config else None
    
    # handle case when input is None (no job_config.yaml file)
    if not input_job_config:
        logger.debug("No input_job_config, creating a new PositronJob()")
        return_config = PositronJob()
    else:
        return_config = copy.deepcopy(input_job_config)
    
    logger.debug(f"return_config: {id(return_config)} created from input_job_config: {id(input_job_config)}")

     # 
     # Name
     #
    if return_config.name:
        if name_arg:
            logger.debug(f"Overriding return_config.name: {return_config.name} <-- name_arg: {name_arg}")
            return_config.name = name_arg
            return_config.name_selection = "overridden by argument"
    else:
        if name_arg:
            logger.debug(f"Setting return_config.name: <-- name_arg: {name_arg}")
            return_config.name = name_arg
            return_config.name_selection = args_job_config.name_selection
    #
    # Remote Python Version
    #
    if return_config.rpv:
        # set in the .yaml?
        if rpv_arg:
            logger.debug(f"Overriding return_config.rpv: {return_config.rpv} <-- rpv_arg: {rpv_arg}")
            return_config.rpv = rpv_arg
            return_config.rpv_selection = "overridden by argument"
    else:
        if rpv_arg:
            logger.debug(f"Setting return_config.rpv: <-- rpv_arg: {rpv_arg}")
            return_config.rpv = rpv_arg
            return_config.rpv_selection = "set by argument"
        else:
            return_config.rpv = "local"
    #
    # Include Local Dir
    #
    if return_config.include_local_dir:
        if include_local_dir_arg:
            logger.debug(f"Overriding return_config.include_local_dir: {return_config.include_local_dir} <-- include_local_dir_arg: {include_local_dir_arg}")
            return_config.include_local_dir = include_local_dir_arg
            return_config.include_local_dir_selection = "overridden by argument"
    else:
        if include_local_dir_arg:
            logger.debug(f"Setting return_config.include_local_dir: <-- include_local_dir_arg: {include_local_dir_arg}")
            return_config.include_local_dir = include_local_dir_arg
            return_config.include_local_dir_selection = args_job_config.include_local_dir_selection
    
    #
    # Mode
    #
    if return_config.mode:
        if mode_arg:
            logger.debug(f"overriding return_config.mode: {return_config.mode} <-- mode_arg: {mode_arg}")
            return_config.mode = mode_arg

    else:
        if mode_arg:
            logger.debug(f"setting return_config.mode_arg: <-- mode_arg: {mode_arg}")
            return_config.mode = mode_arg

    #
    # Dependencies
    #
    if return_config.dependencies:
        if dependencies_arg:
            logger.debug(f"overriding return_config.dependencies: {return_config.dependencies} <-- dependencies_arg: {dependencies_arg}")
            return_config.dependencies = dependencies_arg
            return_config.dep_selection = "overridden by argument"
    else:
        if dependencies_arg:
            logger.debug(f"setting return_config.dependencies: <-- dependencies_arg: {dependencies_arg}")
            return_config.dependencies = dependencies_arg
            return_config.dep_selection = "set by argument"

    # 
    # Funding Group
    #
    if return_config.funding_group_id:
        if funding_arg:
            logger.debug(f"Setting or overriding return_config.funding_group_id: {return_config.funding_group_id} <-- funding_arg: {funding_arg}")
            return_config.funding_group_id = funding_arg
            return_config.funding_selection = "overridden by argument"
        else:
            logger.debug(f"return_config.funding_group_id: already set: {return_config.funding_group_id}")
            # ok the correct funding group is already in return_config.funding_group_id
            pass
    else:
        if funding_arg:
            logger.debug(f"Setting return_config.funding_group_id: {return_config.funding_group_id} <-- funding_arg: {funding_arg}")
            return_config.funding_group_id = funding_arg
            return_config.funding_selection = "setting by argument"
        else:
            # no funding group id in job_config.yaml or argument
            logger.debug(f"No funding group id in job_config.yaml or argument")
        
    # pre-fetch the funding sources info and def env
    fs = list_funding_sources()
    if len(fs) == 0:
        logger.debug("No funding sources found.")
        return None

    def_env_id = None
    # Have we identified a funding group yet?
    if return_config.funding_group_id:
        logger.debug(f"Yes, we have a funding group: {return_config.funding_group_id}")
        if not return_config.environment_id:
            for _, val in fs.items():
                if (val[FS_ID] == return_config.funding_group_id):
                    def_env_id = val.get(FS_DEF_ENV_ID)
            if not def_env_id:
                console.print(f"[bold red]Can't get the default environment for the funding group: {return_config.funding_group_id}")
                return None   
            if not return_config.environment_id:
                logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
                return_config.environment_id = def_env_id
                return_config.environment_selection = "default"
        else:
            if not environment_arg and (not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id)):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not in the funding group: {return_config.funding_group_id}")
                return None
    else: 
        # No funding group id, let's get the user's personal and defaut environment
        logger.debug("Still no funding group id, let's get the user's personal one.")
        for _, val in fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return_config.funding_group_id = val.get(FS_ID)
                def_env_id = val.get(FS_DEF_ENV_ID)
                logger.debug(f"Setting return_config.funding_group_id to PERSONAL: {return_config.funding_group_id}, fetched def_env_id: {def_env_id}")
        if return_config.funding_group_id:
            return_config.funding_selection = "default"
        else:
            console.print("[bold red]Can't get the default funding source (personal).")
            return None
        if not def_env_id:
            console.print("[bold red]Can't get the default environment for the personal funding source.")
            return None
        # the user specified an environment in the job_config.yaml file
        if return_config.environment_id:
            # just for fun, lets check if its in the funding group
            if not environment_arg and (not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id)):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not in the funding group: {return_config.funding_group_id}")
                return None
        else:
            logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
            return_config.environment_id = def_env_id
            return_config.environment_selection = "default from Funding Source"
        
    # At this point we should have a funding group id and a default environment id
    if environment_arg:
        # check to make certain the environment argument is in the funding group
        if not _env_in_funding_group(environment_arg, return_config.funding_group_id):
            console.print(f"[bold red] Sorry, the environment you passed as an argument: {environment_arg} is not in the funding group: {return_config.funding_group_id}")
            return None
        logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- environment_arg: {environment_arg}")
        return_config.environment_id = environment_arg
        return_config.environment_selection = "overridden by argument"

    # Use configured image or auto-select
    if return_config.image and not image_arg:
        logger.debug(f'Using the image: {return_config.image}. To override, please use the --image option.')
        return_config.image_selection = f"from {return_config.filename}"
    elif return_config.image and image_arg:
        logger.debug(f'overriding image: {return_config.image}, setting image to: {image_arg}')
        return_config.image = image_arg
        return_config.image_selection = "overridden by argument"
    elif not return_config.image and image_arg:
        logger.debug(f'setting image: {return_config.image}, setting image to: {image_arg}')
        return_config.image = image_arg
        return_config.image_selection = "set by argument"
    elif not return_config.image:
        logger.debug(f'No image found, setting to auto-select')
        return_config.image = "auto-select"
        
    logger.debug(return_config.to_string("Exiting _merge_from_yaml_and_args()"))

    return return_config

def _env_in_funding_group(env_id, funding_group_id):
    """ sanity check a environment id is in the funding group """
    
    if env_id == "auto-select":
        return True
    try :
        envs = list_environments(funding_group_id)
        for _, val in envs.items():
            if (val[FS_ID] == env_id):
                return True
        return False
    except Exception as e:
        console.print(f"[bold red]Error: {str(e)}")
        return False

# if main then load and validate
''''
if __name__ == "__main__":
    job_config = load_job_config()
    if job_config:
        job_config.validate_values()
        print(job_config)
'''



import time
import os
from rich import box
from rich.text import Text
from rich.panel import Panel
from rich.console import Group
from rich.tree import Tree
from common.config import PositronJob
from common.console import console, ROBBIE_ORANGE, ROBBIE_BLUE
from common.exceptions import RemoteCallException
from common.logging_config import logger
from common.enums import JobRunType
from positron_job_runner.runtime_environment_manager import (
    _running_in_conda, 
    CONDA_MODE,
    PYTHON_MODE,
    GENERIC_MODE
)
from common.utils import _current_python_version

workspace_file_tree = Tree("Files to be copied to remote machine:", style=ROBBIE_BLUE)

def print_robbie_configuration_banner(job: dict, job_config: PositronJob):

    # print(f"print_robbie_configuration_banner cfg: {job_config.to_string()}")

    text = Text(end="")

    if job_config.name:
        text.append("Run Name: ", style=ROBBIE_BLUE)
        text.append(f"{job_config.name}\n")

    text.append("Run Type: ", style=ROBBIE_BLUE)
    if job_config.job_type == JobRunType.BASH_COMMAND_RUNNER:
        text.append("/bin/bash command runner\n")
        text.append("Remote Mode: ", style=ROBBIE_BLUE)
        if job_config.mode == CONDA_MODE:
            if job_config.mode_selection:
                text.append(f"Conda - {job_config.mode_selection}\n")
            else:
                text.append(f"Conda\n")
            text.append("Remote Python Version: ", style=ROBBIE_BLUE)
            text.append(f"{job_config.rpv} - {job_config.rpv_selection}\n")  
            text.append("Environment file: ", style=ROBBIE_BLUE)
            if not job_config.dependencies:
                text.append("None specified\n")
            else:
                text.append(f"{job_config.dependencies} - {job_config.dep_selection}\n")
        elif job_config.mode == PYTHON_MODE:
            if job_config.mode_selection:
                text.append(f"Python - {job_config.mode_selection}\n")
            else:
                text.append(f"Python\n")
            text.append("Remote Python Version: ", style=ROBBIE_BLUE)
            text.append(f"{job_config.rpv} - {job_config.rpv_selection}\n")  
            text.append("Dependency file: ", style=ROBBIE_BLUE)
            if not job_config.dependencies:
                text.append("None specified\n")
            else:
                text.append(f"{job_config.dependencies} - {job_config.dep_selection}\n")
        elif job_config.mode == GENERIC_MODE:
            text.append(f"Generic\n")
        else:
            console.print(f"[bold red]Unknown command runner mode: {job_config.mode}")
            return
    elif job_config.job_type == JobRunType.REMOTE_FUNCTION_CALL:
        text.append("Remote Function\n")
        text.append("Remote Mode: ", style=ROBBIE_BLUE)
        if job_config.mode == CONDA_MODE:
            if job_config.mode_selection:
                text.append(f"Conda - {job_config.mode_selection}\n")
            else:
                text.append(f"Conda\n")
            text.append("Remote Python Version: ", style=ROBBIE_BLUE)
            text.append(f"{job_config.rpv} - {job_config.rpv_selection}\n")  
            text.append("Environment file: ", style=ROBBIE_BLUE)
            if not job_config.dependencies:
                text.append("None specified\n")
            else:
                text.append(f"{job_config.dependencies} - {job_config.dep_selection}\n")
        elif job_config.mode == PYTHON_MODE:
            if job_config.mode_selection:
                text.append(f"Python - {job_config.mode_selection}\n")
            else:
                text.append(f"Python\n")
            text.append("Remote Python Version: ", style=ROBBIE_BLUE)
            text.append(f"{job_config.rpv} - {job_config.rpv_selection}\n")  
            text.append("Dependency file: ", style=ROBBIE_BLUE)
            if not job_config.dependencies:
                text.append("None specified\n")
            else:
                text.append(f"{job_config.dependencies} - {job_config.dep_selection}\n")
    else:
        console.print(f"[bold red]Unknown job type: {job_config.job_type}")
        return

    text.append("Robbie Python SDK Version: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.robbie_sdk_version}\n")
    
    text.append("Funding Source: ", style=ROBBIE_BLUE)
    if not job_config.funding_selection:
        job_config.funding_selection = "Default"
    if job_config.verbose:
        text.append(f"{job['fundingGroupName']} ({job['fundingGroupId']}) - {job_config.funding_selection}\n")
    else:
        text.append(f"{job['fundingGroupName']} - {job_config.funding_selection}\n")
        
    text.append("Hardware: ", style=ROBBIE_BLUE)
    if not job_config.environment_selection:
        job_config.environment_selection = "Default"
    if job_config.verbose:
        text.append(f"{job['environmentName']} ({job['environmentId']}) - {job_config.environment_selection}\n")
    else:
        text.append(f"{job['environmentName']} - {job_config.environment_selection}\n")

    if job_config.cluster:
        text.append("Cluster: ", style=ROBBIE_BLUE)
        text.append(f"{job_config.cluster}\n")
        
    text.append("Image: ", style=ROBBIE_BLUE)
    if not job_config.image_selection:
        job_config.image_selection = "Default" 
    text.append(f"{job['imageName']} - {job_config.image_selection}\n")

    # max token limit
    if job["maxUsableTokens"]:
        text.append("Max Token Consumption: ", style=ROBBIE_BLUE)
        text.append(f"{job['maxUsableTokens']}\n")
    elif job_config.verbose:
        text.append("Max Token Consumption: ", style=ROBBIE_BLUE)
        text.append("Not specified\n")

    # max time limit
    if job["maxExecutionMinutes"]:
        text.append("Max Execution Time (minutes): ", style=ROBBIE_BLUE)
        text.append(f"{job['maxExecutionMinutes']}\n")
    elif job_config.verbose:
        text.append("Max Execution Time (minutes): ", style=ROBBIE_BLUE)
        text.append("Not specified\n")

    # environment variables
    if job_config.env:
        text.append("Environment Variables: ", style=ROBBIE_BLUE)
        text.append(f"{job_config.env}\n")
    elif job_config.verbose:
        text.append("Environment Variables: ", style=ROBBIE_BLUE)
        text.append(" Not specified\n")

    # shell commands
    if job_config.job_type == JobRunType.REMOTE_FUNCTION_CALL:
        if job_config.commands:
            text.append("Pre-exection shell commands:  \n", style=ROBBIE_BLUE)
            for cmd in job_config.commands:
                text.append(f' - {cmd}\n')
        elif job_config.verbose:
            text.append("No pre-exec shell commands specified\n")
    elif job_config.job_type == JobRunType.BASH_COMMAND_RUNNER:
        if job_config.commands:
            text.append("Shell commands:  \n", style=ROBBIE_BLUE)
            for cmd in job_config.commands:
                text.append(f' - {cmd}\n')
        elif job_config.verbose:
            text.append("No shell commands specified\n")
    else:
        console.print(f"[bold red]Unknown job type: {job_config.job_type}")
        return
    
    # workspace directory
    text.append(f"Copy local CWD to remote machine: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.include_local_dir}\n") 

    if job_config.include_local_dir:
        # custom file filter
        if job_config.custom_file_filter:
            text.append("Custom file filters (exclude):\n", style=ROBBIE_BLUE)
            for f in job_config.custom_file_filter:
                text.append(f" - {f}\n")

        group = Group(text, workspace_file_tree, fit=False)
    else:
        group = Group(text, fit=False)

    console.print(Panel(
        group,
        box=box.ROUNDED,
        # padding=(1, 2),
        title = Text(f"Robbie Run Configuration ({job['tokenRatePerHour']} tokens/hour)", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"========== Robbie Run Configuration ({job['tokenRatePerHour']} tokens/hour) ========== \n{text}")

def print_job_details_banner(job: dict):
    ## print job details
    text = Text()
    text.append("Run Name: ", style=ROBBIE_BLUE)
    text.append(f"{job['name']}\n")
            
    text.append("Run ID: ", style=ROBBIE_BLUE)
    text.append(f"{job['id']}\n")
        
    text.append("Start Time: ", style=ROBBIE_BLUE)
    text.append(f"{time.asctime()}")

    console.print(Panel(
        text,
        box=box.ROUNDED,
        title=Text("Run Details", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"\n========== Run Details ========== \n{text}")
    
# prints a rich job completion banner
def print_job_complete_banner(job: dict, start):
    ## print job details
    text = Text()
    text.append("Job Name: ", style=ROBBIE_BLUE)
    text.append(f"{job['name']}\n")
            
    text.append("Total time: ", style=ROBBIE_BLUE)
    text.append(f"{time.perf_counter() - start:.2f} seconds.\n")
        
    text.append("Tokens consumed: ", style=ROBBIE_BLUE)
    text.append(f"{job['tokensUsed']}\n")
        
    text.append("RESULT: ")
    if(job['status'] == "complete"):
        text.append(f"Success", style="green")
    else:
        text.append(f"{job['status']}", style="red")
                
    console.print(Panel(
        text,
        box=box.ROUNDED,
        # padding=(1, 2),
        title=Text("Run Complete", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"========== Run Complete ========== \n{text}")    

def print_known_error(e: RemoteCallException):
    logger.debug(e, exc_info=True)
    console.print(f"[red]An error has occurred: {e.user_friendly_message}[/red]")
    if e.additional_help:
        console.print(f"[yellow]{e.additional_help}[/yellow]")



import uuid
import typer
from rich.prompt import Confirm
from typing import Optional
from typing_extensions import Annotated
from common.api.list_jobs import list_jobs
from common.api.delete_job import delete_job
from common.api.get_job import get_job
from common.observability.main import track_command_usage
from common.console import console
from common.utils import is_valid_uuid
from cli.cmds.login import login


@track_command_usage("delete")
def delete(
  type: Annotated[str, typer.Argument(help='run')],
  id: Annotated[Optional[str], typer.Argument(help="ID of the run to delete")] = None,
  all: Annotated[Optional[bool], typer.Option("--all", help="Delete all runs")] = None,
  yes: Annotated[Optional[bool], typer.Option("--y", help="Skip confirmation")] = None
):
    """
    Delete a run artifacts

    Usage: robbie delete run <id> | --all

    `id` can be the positional ID returned from get runs or the UUID of the run

    """

    if type != "run":
        console.print("Please provide a valid object type to delete")
        return
    
    if id is not None and all:
        console.print("Please provide a run ID or use --all to delete all runs")
        return
    
    login()

    try:
        # get all the jobs  
        jobs = list_jobs()
        if len(jobs) == 0:
            console.print("No jobs found")
            return

        if all:
            # loop through and delete all runs
            if yes or Confirm.ask("[bold red]Are you sure you want to delete all runs?", default=False):
                for key, val in jobs.items():
                    try:
                        delete_job(val["id"])
                        console.print(f"Successfully deleted run: {key}, name: {val['name']} - uuid: {val['id']}")
                    except Exception as e:
                        console.print(f"Failed to delete run: {key}, name: {val['name']} - uuid: {val['id']}, error: {e}")
            else:
                console.print("No runs were deleted")
            return

        # Now figure out which type of ID we are dealing with

        # Support the ability to kill by UUID too
        if is_valid_uuid(id):
            if yes or Confirm.ask(f"[bold red]Are you sure you want to delete run:{id}?", default=False):
                try:
                    val = get_job(id)
                    delete_job(id)
                    console.print(f"Successfully deleted run: {id}, name: {val['name']} - uuid: {val['id']}")
                    return
                except Exception as e:
                    console.print(f"Failed to delete run: {id}, error: {e}")
                    return
            else:
                console.print("Run not deleted")
                return

        # User provided a short run ID from the `robbie get runs` command
        for key, val in jobs.items():
            if key == id:
                if yes or Confirm.ask(f"[bold red]Are you sure you want to delete run:{id}?", default=False):
                    try:
                        delete_job(val["id"])
                        console.print(f"Successfully deleted run: {id}, name: {val['name']} - uuid: {val['id']}")
                        return
                    except Exception as e:
                        console.print(f"Failed to delete run: {id}, name: {val['name']} - uuid: {val['id']}, error: {e}")
                        return
                else:
                    console.print("Run not deleted")
                    return

        console.print(f"Run {id} not found")
        return
    except Exception as e:
        console.print(f"Error deleting run: {id}, error: {e}")
        return
    


import os
import sys
import typer
import platform
import subprocess
import select
import threading
import boto3
import time
import awscli.clidriver
from awscli.testutils import create_clidriver
from typing import Optional, List
from typing_extensions import Annotated
from common.api.get_pd_auth import get_pd_auth
from common.observability.main import track_command_usage
from common.env_defaults import current
from common.console import console
from cli.cmds.login import login
from common.logging_config import logger


@track_command_usage("pd")
def pd(
  command: Annotated[str, typer.Argument(help='Peristent disk operation: ls, cp, mv, rm, mkdir')],
  arg_1: Annotated[Optional[str], typer.Argument(help='Argument 1')] = None,
  arg_2: Annotated[Optional[str], typer.Argument(help='Argument 2')] = None,
  recursive: Annotated[bool, typer.Option("--recursive", help='Recursively operate on files')] = False,
  exclude: Annotated[List[str], typer.Option("--exclude", help='Exclude specific files/regex')] = None,
  include: Annotated[List[str], typer.Option("--include", help='Includes specific files/regex')] = None,
  dryrun: Annotated[bool, typer.Option("--dryrun", help='Dry run the command')] = False
):
    """
    Persistent disk operations: ls, cp, mv, rm, mkdir
    """
    login()

    # what stage are we using?
    stage = current.name
    logger.debug(f"Current stage: {stage}")

    try:
        result = get_pd_auth()
        identity_id = result['identityId']
        user_email = result['userEmail']
        logger.debug(f"Retrieved Identity ID: {identity_id}")
        logger.debug(f"User Email: {user_email}")
        #print(f"Retrieved Identity ID: {identity_id}")

        # Initialize the Cognito Identity client
        client = boto3.client('cognito-identity', region_name='us-west-2')  # Change region as needed

        logger.debug(f"Getting AWS credentials for identity:{identity_id}")
        response = client.get_credentials_for_identity(
            IdentityId=identity_id
        )
        credentials = response['Credentials']
        if not credentials:
            console.print("Error: Unable to retrieve AWS credentials")
            return

        logger.debug(f"Temporary AWS Credentials:{credentials}") 

        os.environ['AWS_ACCESS_KEY_ID'] = credentials['AccessKeyId']
        os.environ['AWS_SECRET_ACCESS_KEY'] = credentials['SecretKey']
        os.environ['AWS_SESSION_TOKEN'] = credentials['SessionToken']
        os.environ['AWS_DEFAULT_REGION'] ="us-west-2"

        if command not in [ "ls", "cp", "mv", "rm", "mkdir"]:
            console.print("Error: Invalid persistent disk command")
            return

        # users have permission to read/write to their own persistent disk
        pd_base = f"s3://positron-{stage}-workspaces/{user_email}/persistent-job-disk/"

        # are we dry running?
        if dryrun:
            dr = " --dryrun"
        else:
            dr = ""

        # are we do a recursive operation?
        if recursive:
            rec = " --recursive"
        else:
            rec = ""

        # process one or more excludes
        excl = ""
        if exclude:
            for e in exclude:
                excl += f" --exclude {e}"

        # process one or more includes
        incl = ""
        if include:
            for i in include:
                incl += f" --include {i}"   

        # list the contents of the persistent disk/directory
        if command == "ls":
            if arg_1:
                if not arg_1.startswith("pd://"):
                    console.print("Error: rm requires an `pd` location.")
                    return
                arg_1_full = f'{pd_base}{arg_1.removeprefix("pd://")}'
                command = f"aws s3 ls {arg_1_full}{rec}{dr} --human-readable --summarize"
            else:
                command = f"aws s3 ls {pd_base}{rec}{dr} --human-readable --summarize"

        # copy files to/from the persistent disk
        if command == "cp":
            if not arg_1 or not arg_2:
                console.print("Error: cp requires two arguments.")
                return

            # if is an S3 location, expand the path 
            if arg_1.startswith("pd://"):
                arg_1_full = f'{pd_base}{arg_1.removeprefix("pd://")}'
            else:
                arg_1_full = arg_1

            # if is an S3 location, expand the path    
            if arg_2.startswith("pd://"):
                arg_2_full = f'{pd_base}{arg_2.removeprefix("pd://")}'
            else:
                arg_2_full = arg_2

            command = f"aws s3 cp {arg_1_full} {arg_2_full}{incl}{excl}{rec}{dr}"

        # move files to/from the persistent disk
        if command == "mv":
            if not arg_1 or not arg_2:
                console.print("Error: mv requires two arguments.")
                return

            # if is an S3 location, expand the path 
            if arg_1.startswith("pd://"):
                arg_1_full = f'{pd_base}{arg_1.removeprefix("pd://")}'
            else:
                arg_1_full = arg_1

            # if is an S3 location, expand the path    
            if arg_2.startswith("pd://"):
                arg_2_full = f'{pd_base}{arg_2.removeprefix("pd://")}'
            else:
                arg_2_full = arg_2

            command = f"aws s3 mv {arg_1_full} {arg_2_full}{incl}{excl}{rec}{dr}"
        
        # create a new directory on the persistent disk
        if command == "mkdir":
            if not arg_1:
                console.print("Error: mkdir requires one argument.")
                return
            if not arg_1.startswith("pd://"):
                console.print("Error: mkdir requires an `pd://` location.")
                return
            if recursive:
                console.print("Error: mkdir doesn't support the --recursive option.")
                return
            if arg_2:
                console.print("Error: mkdir doesn't a second argument.")
                return
            if include or exclude:
                console.print("Error: mkdir support --include or --exclude options.")
                return

            # just strip off the pd:// prefix
            arg_1_full = arg_1.removeprefix("pd://")
            command = f"aws s3api put-object --bucket positron-{stage}-workspaces --key {user_email}/persistent-job-disk/{arg_1_full}/{dr}"

        if command == "rm":
            if not arg_1:
                console.print("Error: rm requires one argument.")
                return
            if not arg_1.startswith("pd://"):
                console.print("Error: rm requires an a `pd://` location.")
                return
            if arg_2:
                console.print("Error: rm doesn't a second argument.")
                return
            if arg_1.endswith("persistent-job-disk") or arg_1.endswith("persistent-job-disk/"):
                console.print("Error: Cannot delete the root directory.")
                return
            if recursive:
                rec = " --recursive"
            else:
                rec = ""
            
            arg_1_full = f'{pd_base}{arg_1.removeprefix("pd://")}'
            command = f"aws s3 rm {arg_1_full}{incl}{excl}{rec}{dr}"

        if not command:
            console.print("Error: no command to run")
            return

        cmd_list = command.split()
        if not isinstance(cmd_list, list):
            console.print("Error: Command must be a list")
            return
        
        # override sys.argv to pass these arguments to awscli driver (function)
        sys.argv = cmd_list
        logger.debug(f"new sys.argv:{sys.argv}")
        # call the library directly. This eliminates the need to call the awscli command from the shell
        return_code = awscli.clidriver.main()
        if return_code == 0:
            console.print("SUCCESS")
        else:
            console.print(f"ERROR: {return_code}")
        return
    
    except Exception as e:
        console.print(f"Error running command: {e}")
        return






import subprocess
import threading
import select
import os
from typing import Dict


def run_local_docker_container(job_id: str, base_path: str):
    print("run_local_docker_container")
    
    # env = {}
    env: Dict[str, str] = {}
    env['JOB_ID'] = job_id
    if os.getenv('REMOTE_FUNCTION_SECRET_KEY'):
        env['REMOTE_FUNCTION_SECRET_KEY'] = os.getenv('REMOTE_FUNCTION_SECRET_KEY')
    print("env=", env)

    p = subprocess.Popen(
        "env && /usr/local/bin/docker compose up --no-color --build --force-recreate",
        text=True,
        shell=True,
        stdout=subprocess.PIPE, 
        stderr=subprocess.PIPE,
        env=env,
        cwd=f"{base_path}/test/job_runner",
    )
    log_stop_event = threading.Event()
    stdout_thread = threading.Thread(target=logging_thread, args=(p.stdout, "stdout", log_stop_event)); stdout_thread.start()
    stderr_thread = threading.Thread(target=logging_thread, args=(p.stderr, "stderr", log_stop_event)); stderr_thread.start()
    p.wait()
    log_stop_event.set()
    stdout_thread.join()
    stderr_thread.join()
    print("p.returncode=", p.returncode)

def logging_thread(pipe, stream_name: str, stop_event: threading.Event):
    try:
        with pipe:
            while not stop_event.is_set():
                # Calling pipe.readline() will hang the process until a new line is available
                # however, if no new lines are available, we want to check if the thread should stop
                ready, _, _ = select.select([pipe], [], [], 0.1)
                if ready:
                    # TODO: Can I iterate over the pipe and process everything that's ready?
                    line = pipe.readline()
                    if not line:
                        break
                    if line.strip():
                        if stream_name == "stdout":
                            print(line.rstrip())
                        else:
                            print("ERROR: " + line.rstrip())
                else:
                    stop_event.wait(1)

            # Read one last time.
            ready, _, _ = select.select([pipe], [], [], 0.1)
            if ready:
                for line in pipe:
                    if line.strip():
                        if stream_name == "stdout":
                            print(line.rstrip())
                        else:
                            print("ERROR: " + line.rstrip())
            # TODO: Do we need to check one more time for outstanding data in the pipe?
            # logger.debug(f'Logging thread for: {stream_name} stopped')
    except Exception as e:
        print(f'Logging thread for: {stream_name} stopped with exception: {e}')







import os
import re
import typer
import pyfiglet
import platform
from typing_extensions import Annotated
from typing import Optional
from prompt_toolkit import prompt
from prompt_toolkit.styles import Style
from common.config import (
    PositronJob,
    load_job_config, 
    merge_from_yaml_and_args
)
from common.cli_args import args
from common.console import console, ROBBIE_DEFAULT, ROBBIE_BLUE
from common.constants import JOB_CONF_YAML_PATH
from common.logging_config import logger
from common.enums import JobRunType
from common.image_name import get_auto_image_name_and_cluster
from common.observability.main import track_command_usage
from common.api.funding_envs_images import *
from common.api.get_env_rec import (
    get_env_rec,
)
from cli.cmds.download import filename_is_valid
from cli.cmds.login import login
from cli.deploy import command_runner_deploy
from cli.cmds.rec import (
    build_prompt_from_envs,
    find_uuids
)
from cli.interactive import (
    prompt_and_build_positron_job_config,
    get_job_config_yaml_name
)
from common.env_defaults import current
from cli.auto_complete import (
    funding_group_auto_complete, 
    environment_auto_complete, 
    images_auto_complete,
    deps_auto_complete,
    job_config_auto_complete,
    cluster_auto_complete,
)
from positron_job_runner.runtime_environment_manager import (
    RuntimeEnvironmentManager, 
    _build_job_args, 
    _running_in_conda,
    _get_python_version_from_conda_yaml,
    CONDA_MODE,
    PYTHON_MODE,
    GENERIC_MODE
)

#
# Naming convention for the arguments:
#
# `jctr` is the PositronJob object that will be passed to deploy to run the job
# `rpv` is the Python version to run the job on the remote machine
# `deps` is the file containing the dependencies or auto-capture
#

@track_command_usage("run")
def run(
  # the following options control the job configuration
  job_config_arg: Annotated[str, 
                            typer.Option("--f", 
                                        help="Specify the job configuration file.",
                                        autocompletion=job_config_auto_complete)] = "./job_config.yaml",
  name_arg: Annotated[str, typer.Option("--name", help="Name of the run")] = None,
  funding_arg: Annotated[str, 
                         typer.Option("--funding_group_id", 
                                help="Specify the team to run the job on.",
                                autocompletion=funding_group_auto_complete)] = None,
  environment_arg: Annotated[str, 
                             typer.Option("--environment_id", 
                                    help="Specify the hardware to run the job on.",
                                    autocompletion=environment_auto_complete)] = None,
  image_arg: Annotated[str, 
                       typer.Option("--image", 
                            help="Specify the image to run the job on or 'auto-select'.",
                            autocompletion=images_auto_complete)] = None,
  commands: Annotated[Optional[str], typer.Argument(help='Shell command(s)')] = None,
  depfile_arg: Annotated[str, 
                         typer.Option("--deps", 
                                      help='Specify the file containing the deps or auto-capture.',
                                      autocompletion=deps_auto_complete)] = None,
  conda_arg: Annotated[bool, typer.Option("--conda", help='Run on remote machine in a conda environment')] = None,
  python_arg: Annotated[bool, typer.Option("--python", help='Run on remote machine in a Python environment', is_flag=False)] = None,
  python_ver_arg: Annotated[str, typer.Option("--version", help='Specify the Python version to run the job on.')] = None,    
  include_local_dir_arg: Annotated[bool, typer.Option("--include-local-dir", help='Include the working directory in the job.')] = False,

  cluster_arg: Annotated[str, typer.Option("--cluster",
                                        help="Specify the cluster to run the job on.", 
                                        autocompletion=cluster_auto_complete)] = None,

    
  # the following options control how the job is run
  tail: Annotated[bool, typer.Option("--tail", help='Tail the run\'s stdout back to your CLI.')] = False,
  skip_prompts: Annotated[bool, typer.Option("--y", help='Bypass the prompts and execute the run immediately.')] = False,
  interactive: Annotated[bool, typer.Option("--i", help="Interactively choose your run configuration.")] = False,
  download: Annotated[str, typer.Option("--download", help="Download a results <file> to your local machine. Specify 'all' for the complete list.")] = None,
  path: Annotated[str, typer.Option("--path", help="Local directory where the downloaded files will be stored.")] = None,
  create_only: Annotated[bool, typer.Option("--create-only", help="Only create the run, do not execute it. [Robbie internal use only]")] = False,
  local_container: Annotated[str, typer.Option("--local", help="Run the job in a local container <path of source>. [Robbie internal use only]")] = False,
  verbose: Annotated[bool, typer.Option("--v", help="Show advanced details.")] = False,
) -> None:
    """
    Run shell commands as a batch job in Robbie

    There are three possible scenarios:
    1. User types: robbie run - commands are run from the job_config.yaml file
    2. User types: robbie run "command" - commands override the job_config.yaml file
    3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created/overwritten

    In scenerio 1 and 2, user's can pass in funding groups and enviroments as arguments to override the job_config.yaml file.

    """
    ascii_banner = pyfiglet.figlet_format("Robbie")
    console.print(ascii_banner, style='#41a7ff')

    console.print(f"[blue]Platform: [white]{platform.system()}[blue] Stage: [white]{current.name}")
    if platform.system() == "Windows" and os.getenv('CONDA_DEFAULT_ENV'):
        console.print(f"[yellow] Windows conda support is currently experimental.")

    logger.debug(f"""========== run() arguments ==========
    - name_arg: {name_arg}
    - funding_arg: {funding_arg}
    - environment_arg: {environment_arg}
    - image_arg: {image_arg}
    - commands: {commands}
    - depfile_arg: {depfile_arg}
    - conda_arg: {conda_arg}
    - python_arg: {python_arg}
    - include_local_dir_arg: {include_local_dir_arg}
    - verbose: {verbose}
    - tail: {tail}
    - skip_prompts: {skip_prompts}
    - interactive: {interactive}
    - download: {download}
    - path: {path}
    - create_only: {create_only}
    - local_container: {local_container}""")

    # check if the user is logged in
    login()

    console.print("[green]✔[/green] Current working directory: ", os.getcwd(), style=ROBBIE_BLUE)
    # was there a job_config.yaml file in the CWD?

    try:
        global_job_config_yaml = load_job_config(job_config_arg)
        if global_job_config_yaml:
            logger.debug(f"global_job_config_yaml: {global_job_config_yaml.to_string()}")
            console.print(f"[green]✔[/green] Loaded run configuration file: ", end="", style=ROBBIE_BLUE)
            console.print(f"{job_config_arg}", style=ROBBIE_DEFAULT)
            if global_job_config_yaml.job_type != JobRunType.BASH_COMMAND_RUNNER:
                console.print(f"[red]Error, {job_config_arg} file is not configured for bash command runner.")
                return
        else:
            console.print(f"[yellow]{job_config_arg} not found.")
            logger.debug(f"{job_config_arg} not found.")
    except Exception as e:
        print(f'Error loading job configuration! {str(e)}')
        return

    if _running_in_conda():
        console.print(f"[green]✔[/green] Running in conda environment: {os.getenv('CONDA_DEFAULT_ENV')}", style=ROBBIE_BLUE)
    else:
        console.print(f"[green]✔[/green] Running in non-conda environment", style=ROBBIE_BLUE)

    
    if (download and not tail):    
        console.print('[red]Error: The --download option can only be used with the --tail')
        return
    
    if download and not filename_is_valid(download):
        console.print('[red]Error: Please specify a valid file name or "all" to download all files.')
        return
    
    if path and not download:
        console.print('[red]Error: The --path option can only be used with the --download option.')
        return
    
    if path and not os.path.exists(path):
        console.print('[red]Error: The path you specified is not valid.')
        return

    # initialize the argument singleton
    args.init(
        name=name_arg,
        stream_stdout=tail,
        skip_prompts=skip_prompts,
        commands_to_run=commands,
        interactive=interactive,
        create_only=create_only,
        local_container=local_container,
        download=download,
        local_path=path,
    )

    # first-level sanity checks
    if commands and interactive:
        console.print("[red]Sorry: Please specify command line or use the interactive mode.")
        return
    
    if interactive and (funding_arg or environment_arg or image_arg):
        console.print("[red]Sorry: You can't specify funding, environments, or images in interactive mode.")
        return
    
    if conda_arg and interactive:
        console.print("[red]Sorry: You can't specify remote conda mode in interactive mode.")
        return
    
    if python_arg and interactive:
        console.print("[red]Sorry: You can't specify remote Python modde in interactive mode.")
        return
    
    if depfile_arg and interactive:
        console.print("[red]Sorry: You can't specify deps in interactive mode.")
        return
    
    if conda_arg and python_arg:
        console.print("[red]Sorry: You can't specify conda and python modes at the same time.")
        return
    
    if conda_arg and python_ver_arg:
        console.print("[red]Sorry: You can't specify a Python version with conda mode.")
        return
    
    # capture arguments job configuration
    job_config_args = PositronJob()
    if name_arg:
        job_config_args.name = name_arg
    if funding_arg:
        job_config_args.funding_group_id = funding_arg
        job_config_args.funding_selection = "Passed as argument to the run command"
    if environment_arg:
        job_config_args.environment_id = environment_arg
        job_config_args.environment_selection = "Passed as argument to the run command"
    if image_arg:
        job_config_args.image = image_arg
        job_config_args.image_selection = "Passed as argument to the run command"
    if depfile_arg:
        job_config_args.dependencies = depfile_arg
        job_config_args.dep_selection = "Passed as argument to the run command"
    if conda_arg:
        job_config_args.mode = CONDA_MODE
    if python_arg:
        job_config_args.mode = PYTHON_MODE
    if python_ver_arg:
        job_config_args.rpv = python_ver_arg
    if include_local_dir_arg:
        job_config_args.include_local_dir = include_local_dir_arg

    # print(f"job_config_args: {job_config_args.to_string()}")

    """
    There are three possible scenarios:
    1. User types: robbie run - commands are run from the job_config.yaml file
    2. User types: robbie run "command" - commands override the job_config.yaml file
    3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created

    There are three important variables in the code:
    - global_job_config_yaml - this is the PositronJob object created from the job_config.yaml file
    - jctr - this is the PositronJob object that will be passed to deploy to run the job
    - scenerio_1_and_2 - this is a flag that is set for scenerio 1 and 2 to merge the passed arguments into to the jctr object
    """
    scenerio_1_and_2 = False

    # Scenerio 1: User types: "robbie run" - commands are run from the job_config.yaml file
    #
    if not commands and not interactive:
        logger.debug("@@@@@@@@@@@   Scenerio 1: User types: robbie run")
        if not global_job_config_yaml:
            console.print('[red]Error: No job_config.yaml file found or it is invalid. Cannot find commands to run.')
            return
        if not global_job_config_yaml.commands:
            console.print('[red]Error: No commands found in job_config.yaml file.')
            return
        jctr = global_job_config_yaml.deepcopy()
        scenerio_1_and_2 = True

    # Scenerio 2. User types: robbie run "command" - commands override the job_config.yaml file if it exists
    #
    if commands:
        logger.debug("@@@@@@@@@@ Scenerio 2: User types: robbie run 'command'")

        # Is there a job_config.yaml file?
        if global_job_config_yaml:
            if global_job_config_yaml.commands:
                console.print('[yellow] Overriding commands in existing job_config.yaml file.')
            jctr = global_job_config_yaml.deepcopy()
        else:
            jctr = PositronJob()
        
        # Add the commands to the job_config
        jctr.commands = []
        jctr.commands.append(commands)

        scenerio_1_and_2 = True
        
    # now lets deal with the arguments for the first two scenarios
    if scenerio_1_and_2:

        logger.debug(f"scenerio_1_and_2 - jctr: {jctr.to_string()}")
        # Determine the version of Python to use for the job on the remote machine
        # Normally, the user doesnt' touch this and the Python version is picked up from the local environment by default

        # other arguments are merged into the jctr object
        logger.debug("Adding arguments to job_config")
        jctr = merge_from_yaml_and_args(jctr, job_config_args)
        if not jctr:
            logger.debug('[red]Error: Unable to merge yaml and arguments.')
            return
     
    # Scenerio 3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created or modified
    if interactive:
        logger.debug("@@@@@@@@@ Scenerio 3: User types: robbie run --i")

        # lets prompt the user 
        console.print(f"\nPlease follow the prompts to configure your run ([{ROBBIE_DEFAULT}][] = default[/{ROBBIE_DEFAULT}], <tab> for menu, contol-c to exit):", style=ROBBIE_BLUE)
        captured_job_config = prompt_and_build_positron_job_config(
            cmd_runner_dash_i = True,
            job_config_yaml=global_job_config_yaml, 
        )
        if captured_job_config == None:
            console.print(f"[red]Sorry, failed to create a file {JOB_CONF_YAML_PATH}")
            return
        
        logger.debug(f"Interactive cfg: {captured_job_config.python_job.to_string()}")
        # print(f"Interactive cfg: {captured_job_config.python_job.to_string()}")

        if captured_job_config.python_job.commands == None:
            console.print("[red]Error: You did not specify any commands to run.")
            return
        
        filename = get_job_config_yaml_name()
        if filename:
            captured_job_config.write_to_file(filename=filename)
            jctr = captured_job_config.python_job.deepcopy()

     # verbose logs
    if verbose:
        jctr.verbose = True

    # 
    #  Sort out the mode based on deps and Python environment.
    #
    # - CONDA_MODE - Remote runs as a mirror to the local conda environment (Python version and deps)
    # - PYTHON_MODE - Remote runs as a mirror to the local Python environment (Python version and deps)
    # - GENERIC_MODE - Remote runs whatever we specify (Python version and deps)
    #
    # Deal with no mode specified in the job_config.yaml file or the arguments
    # print(f"jctr.mode: {jctr.mode}")
    
    # Note used - if mode wasn't set, his code attempts to infer the mode based on the deps
    
    # Was a mode ever set in job_config.yaml or args, if not default to GENERIC
    if not jctr.mode:
        jctr.mode = GENERIC_MODE

    console.print("Robbie is analyzing run information...", style=ROBBIE_BLUE) 

    if jctr.mode == CONDA_MODE:
        # you want to run in a conda environment
        logger.debug(f"Conda mode: jctr.dependencies: {jctr.dependencies}")
        if not (jctr.dependencies and 
          (jctr.dependencies == "auto-capture" or
            jctr.dependencies.endswith(".yml") or 
            jctr.dependencies.endswith(".yaml"))):
            console.print("[red]Error: You specified 'conda' mode but did not provide a valid conda environment file or specify auto-capture.")
            return
        
        # Set python version based on the environment .yaml
        if (jctr.dependencies and 
            (jctr.dependencies.endswith('.yaml') or 
             jctr.dependencies.endswith(".yml"))
        ):
            yaml_py_vers = _get_python_version_from_conda_yaml(jctr.dependencies)
            if yaml_py_vers:
                # deal with possible 3-digit versions (e.g. 3.10.15)
                yaml_py_vers = convert_to_two_digit_version(yaml_py_vers)
                # console.print(f"Python version: {yaml_py_vers} detected in conda yaml file: {jctr.dependencies}")
                jctr.rpv = yaml_py_vers
                jctr.rpv_selection = f"detected in conda environment file: {jctr.dependencies}"
            else:
                # No python in the .yaml, default to 3.10 for image selection
                jctr.rpv = "3.10"
                jctr.rpv_selection = "default for image selection"

        elif jctr.dependencies == "auto-capture":
            # if we are auto-capturing get the python version from the conda env
            if _running_in_conda():
                 # console.print("Auto-capturing Python version from conda environment....")
                conda_py_vers = RuntimeEnvironmentManager()._python_version_in_conda_env(os.getenv('CONDA_DEFAULT_ENV'))
                 # console.print(f"Python version: {conda_py_vers} detected in environment: {os.getenv('CONDA_DEFAULT_ENV')}")
                if conda_py_vers:
                    jctr.rpv = conda_py_vers
                    jctr.rpv_selection = f"detected in conda environment: {os.getenv('CONDA_DEFAULT_ENV')}"
                    console.print(f"[green]✔[/green] Determined Python version: {jctr.rpv} from conda environment: {os.getenv('CONDA_DEFAULT_ENV')}", style=ROBBIE_BLUE)
                else:
                    jctr.rpv = "3.10"
                    jctr.rpv_selection = "default for image selection"
            else:
                console.print(f"[red]Error: You are not running in a conda environment, please start over and select a .yaml file.")
                return
        else:
            console.print(f"[red]Error: You must select a valid conda environment or file.")
            return
        
    elif jctr.mode == PYTHON_MODE:
        # you want to run non-conda python
        logger.debug(f"Python mode: jctr.dependencies: {jctr.dependencies}, jctr.dep_selection: {jctr.dep_selection}")
        # python version not set
        if not jctr.rpv:
            console.print("[red]Error: PYTHON_MODE, rpv should be set.")
            return
        elif jctr.rpv == "local":
            jctr.rpv = RuntimeEnvironmentManager()._current_python_version()
            jctr.rpv_selection = "local environment"
            console.print(f"[green]✔[/green] Determined local Python version: {jctr.rpv}", style=ROBBIE_BLUE)
        if jctr.dependencies:
            if jctr.dependencies == "auto-capture":
                if _running_in_conda():
                    console.print("[red]Error: You specified 'python' mode auto-capture in a conda environment.")
                    return
                if jctr.rpv != "local" and jctr.rpv != RuntimeEnvironmentManager()._current_python_version():
                    console.print("[red]Error: You can't auto-capture deps and specify a Python version other than the local one.")
                    return
            if not (jctr.dependencies in ["auto-capture", "requirements.txt", "none"]):
                console.print("[red]Error: You specified 'python' mode but did not specify a valid requirements.txt file or auto-capture.")
                return
        # else:
        #    console.print("[red]Error: You specified 'python' mode but did not provide a valid requirements.txt file or specify auto-capture.")
        #    return
    elif jctr.mode == GENERIC_MODE:
        # setting this purely to help with image selection, nothing else
        if not jctr.rpv:
            console.print("[red]Error: GENERIC_MODE, rpv should be set.")
            return
        elif jctr.rpv == "local":
            jctr.rpv = RuntimeEnvironmentManager()._current_python_version()
            jctr.rpv_selection = "local environment"
    else:
        console.print("[red]Error: Invalid mode. Please specify 'conda', 'python', or 'generic'.")
        return

    # 
    # Dependencies - Handle auto capturing deps 
    # 
    if jctr.dependencies == "auto-capture":
        jctr.dep_selection = "auto-captured"
    
    if jctr.dependencies == "none":
        jctr.dep_selection = "Not specified"
        jctr.dependencies = None

    # snapshot() does a couple of things: 1) verifies that a passed file exists; 2) auto-captures deps if needed
    try:
        jctr.dependencies = RuntimeEnvironmentManager().snapshot(jctr.dependencies)
    except Exception as e:
        console.print(f"[red]Error: {e}")
        return
    console.print("[green]✔[/green] Dependencies and environment captured.", style=ROBBIE_BLUE)

    #
    # Auto-detect the environment
    #
    if jctr.funding_group_id:
        if jctr.environment_id == "auto-select":
            # if there a python file we can analyze
            python_file = None
            for cmd in jctr.commands:
                python_file = find_python_file_name(cmd)
                if python_file:
                    break
            if not python_file:
                console.print(f"[red]Error: You selected `auto-select` but we cannot find a Python file in the commands.")
                return

            retun_env = None
            try:
                console.print(f"Using AI to securely determine the best hardware for your workload...", style=ROBBIE_BLUE)
                system_prompt, human_prompt = build_prompt_from_envs(jctr.funding_group_id, python_file, cluster_arg)
                if human_prompt == "":
                    console.print(f"[red]Error - no files to analyze.")
                    return
                return_env = get_env_rec(
                    system_prompt=system_prompt,
                    human_prompt=human_prompt
                )
            except Exception as e:
                console.print(f"Error getting environment recommendation: {str(e)}")
                return

            console.print(f"Analysis Complete!", style=ROBBIE_BLUE)

            uuid = find_uuids(return_env.get("response"))
            if len(uuid)== 1:
                env_id = find_uuids(return_env.get("response"))[0]
            else:
                console.print(f"Error: Could not find a correct number of UUIDs in the response.")
                return
            console.print(f"[green]✔[/green] Recommanded Environment: {env_id}", style=ROBBIE_BLUE)
            jctr.environment_id = env_id
            jctr.environment_selection = "auto-selected using AI (experimental)"

    #
    # Image selection
    #
    if jctr.image == "auto-select":
        jctr.image, jctr.cluster = get_auto_image_name_and_cluster(
            jctr.funding_group_id, 
            jctr.environment_id,
            jctr.rpv
        )
        jctr.image_selection = "auto-selected"
        # console.print('Auto-selecting image:', style=ROBBIE_BLUE, end="")
        # console.print(f'{jctr.image} on {jctr.cluster} cluster', style=ROBBIE_DEFAULT)
        console.print("[green]✔[/green] Automatically selected image", style=ROBBIE_BLUE)

    # this will be psssed to the job runner
    args.job_args = _build_job_args(
        mode = jctr.mode,
        rpv = jctr.rpv,
        dependencies = jctr.dependencies,
    )
    logger.debug(f"job_args: {args.job_args}")

    jctr.job_type = JobRunType.BASH_COMMAND_RUNNER

    logger.debug("Calling command_runner_deploy")
    command_runner_deploy(jctr)

def _remove_package_from_requirements(package_name, file_path='./requirements.txt'):
    """ this is hack when the robbie package gets auto-added to the requirements.txt file in the auto-capture deps mode """
    # Read the file
    with open(file_path, 'r') as file:
        lines = file.readlines()

    # Write back all lines except the one with the package name
    with open(file_path, 'w') as file:
        for line in lines:
            if not line.strip().startswith(package_name):
                file.write(line)

def convert_to_two_digit_version(version: str) -> str:
    """
    Converts a three-digit version number (e.g., 1.2.3) to a two-digit version (e.g., 1.2).
    """
    parts = version.split(".")
    # just return the version if it's already two digits
    if len(parts) == 2:
        return version
    elif len(parts) == 3:
        return f"{parts[0]}.{parts[1]}"
    else:
        raise ValueError(f"Invalid version format: {version}")


def find_python_file_name(text):
    """
    Checks if a string contains a Python file name and returns the file name if found.

    Args:
        text: The string to search within.

    Returns:
        The Python file name (including the .py extension) if found, otherwise None.
    """
    match = re.search(r"(\w+)\.py\b", text)
    if match:
        return match.group(0)
    return None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Helper classes that interact with SageMaker Training service."""
from __future__ import absolute_import
import dataclasses

import os
import shutil

from typing import Callable, List, Optional, Union
from urllib.parse import urlparse

from positron_job_runner.runtime_environment_manager import (
    _running_in_conda,
    _get_python_version_from_conda_yaml
)
from common.custom_file_filter import (
    CustomFileFilter,
    copy_workdir
)
from common.config import PositronJob
from common.logging_config import logger
from common.utils import ( 
    _tmpdir
)
from common.print_banner import workspace_file_tree
from common.console import console, ROBBIE_YELLOW

from common.aws.s3 import S3Uploader

# runtime script names
BOOTSTRAP_SCRIPT_NAME = "bootstrap_runtime_environment.py"
ENTRYPOINT_SCRIPT_NAME = "job_driver.sh"
PRE_EXECUTION_SCRIPT_NAME = "pre_exec.sh"
RUNTIME_MANAGER_SCRIPT_NAME = "runtime_environment_manager.py"
SPARK_APP_SCRIPT_NAME = "spark_app.py"

# training channel names
RUNTIME_SCRIPTS_CHANNEL_NAME = "sagemaker_remote_function_bootstrap"
REMOTE_FUNCTION_WORKSPACE = "sm_rf_user_ws"
JOB_REMOTE_FUNCTION_WORKSPACE = "workspace"
SCRIPT_AND_DEPENDENCIES_CHANNEL_NAME = "pre_exec_script_and_deps"

JOBS_CONTAINER_ENTRYPOINT = [
    "/bin/bash",
    f"/opt/ml/input/data/{RUNTIME_SCRIPTS_CHANNEL_NAME}/{ENTRYPOINT_SCRIPT_NAME}",
]

ENTRYPOINT_SCRIPT = f"""
#!/bin/bash

# Entry point for bootstrapping runtime environment and invoking remote function

set -eu

PERSISTENT_CACHE_DIR=${{SAGEMAKER_MANAGED_WARMPOOL_CACHE_DIRECTORY:-/opt/ml/cache}}
export CONDA_PKGS_DIRS=${{PERSISTENT_CACHE_DIR}}/sm_remotefunction_user_deps_cache/conda/pkgs
printf "INFO: CONDA_PKGS_DIRS is set to '$CONDA_PKGS_DIRS'\\n"
export PIP_CACHE_DIR=${{PERSISTENT_CACHE_DIR}}/sm_remotefunction_user_deps_cache/pip
printf "INFO: PIP_CACHE_DIR is set to '$PIP_CACHE_DIR'\\n"


printf "INFO: Bootstraping runtime environment.\\n"
python /opt/ml/input/data/{RUNTIME_SCRIPTS_CHANNEL_NAME}/{BOOTSTRAP_SCRIPT_NAME} "$@"

if [ -d {JOB_REMOTE_FUNCTION_WORKSPACE} ]
then
    if [ -f "remote_function_conda_env.txt" ]
    then
        cp remote_function_conda_env.txt {JOB_REMOTE_FUNCTION_WORKSPACE}/remote_function_conda_env.txt
    fi
    printf "INFO: Changing workspace to {JOB_REMOTE_FUNCTION_WORKSPACE}.\\n"
    cd {JOB_REMOTE_FUNCTION_WORKSPACE}
fi

if [ -f "remote_function_conda_env.txt" ]
then
    conda_env=$(cat remote_function_conda_env.txt)

    if which mamba >/dev/null; then
        conda_exe="mamba"
    else
        conda_exe="conda"
    fi

    printf "INFO: Invoking remote function inside conda environment: $conda_env.\\n"
    $conda_exe run -n $conda_env python -m sagemaker.remote_function.invoke_function "$@"
else
    printf "INFO: No conda env provided. Invoking remote function\\n"
    python -m sagemaker.remote_function.invoke_function "$@"
fi
"""

ENTRYPOINT_TORCHRUN_SCRIPT = f"""
#!/bin/bash

# Entry point for bootstrapping runtime environment and invoking remote function with torchrun

set -eu

PERSISTENT_CACHE_DIR=${{SAGEMAKER_MANAGED_WARMPOOL_CACHE_DIRECTORY:-/opt/ml/cache}}
export CONDA_PKGS_DIRS=${{PERSISTENT_CACHE_DIR}}/sm_remotefunction_user_deps_cache/conda/pkgs
printf "INFO: CONDA_PKGS_DIRS is set to '$CONDA_PKGS_DIRS'\\n"
export PIP_CACHE_DIR=${{PERSISTENT_CACHE_DIR}}/sm_remotefunction_user_deps_cache/pip
printf "INFO: PIP_CACHE_DIR is set to '$PIP_CACHE_DIR'\\n"


printf "INFO: Bootstraping runtime environment.\\n"
python /opt/ml/input/data/{RUNTIME_SCRIPTS_CHANNEL_NAME}/{BOOTSTRAP_SCRIPT_NAME} "$@"

if [ -d {JOB_REMOTE_FUNCTION_WORKSPACE} ]
then
    if [ -f "remote_function_conda_env.txt" ]
    then
        cp remote_function_conda_env.txt {JOB_REMOTE_FUNCTION_WORKSPACE}/remote_function_conda_env.txt
    fi
    printf "INFO: Changing workspace to {JOB_REMOTE_FUNCTION_WORKSPACE}.\\n"
    cd {JOB_REMOTE_FUNCTION_WORKSPACE}
fi

if [ -f "remote_function_conda_env.txt" ]
then
    conda_env=$(cat remote_function_conda_env.txt)

    if which mamba >/dev/null; then
        conda_exe="mamba"
    else
        conda_exe="conda"
    fi

    printf "INFO: Invoking remote function with torchrun inside conda environment: $conda_env.\\n"
    $conda_exe run -n $conda_env torchrun --nproc_per_node $NPROC_PER_NODE \
    -m sagemaker.remote_function.invoke_function "$@"
else
    printf "INFO: No conda env provided. Invoking remote function with torchrun\\n"
    torchrun --nproc_per_node $NPROC_PER_NODE -m sagemaker.remote_function.invoke_function "$@"
fi
"""

def _prepare_and_upload_runtime_scripts(
    use_torchrun: bool = False,
    nproc_per_node: int = 1,
):
    """Copy runtime scripts to a folder and upload to S3.

    In case of remote function, s3_base_uri is s3_root_uri + function_name.
    In case of pipeline, s3_base_uri is s3_root_uri + pipeline_name. The runtime scripts are
    uploaded only once per pipeline.

    Args:
        spark_config (SparkConfig): remote Spark job configurations.

        s3_base_uri (str): S3 location that the runtime scripts will be uploaded to.

        s3_kms_key (str): kms key used to encrypt the files uploaded to S3.

        sagemaker_session (str): SageMaker boto client session.

        use_torchrun (bool): Whether to use torchrun or not.

        nproc_per_node (int): Number of processes per node.
    """

    with _tmpdir() as bootstrap_scripts:

        # write entrypoint script to tmpdir
        entrypoint_script_path = os.path.join(bootstrap_scripts, ENTRYPOINT_SCRIPT_NAME)
        entry_point_script = ENTRYPOINT_SCRIPT

        if use_torchrun:
            entry_point_script = ENTRYPOINT_TORCHRUN_SCRIPT
            entry_point_script = entry_point_script.replace("$NPROC_PER_NODE", str(nproc_per_node))

        with open(entrypoint_script_path, "w", newline="\n") as file:
            file.writelines(entry_point_script)

        bootstrap_script_path = os.path.join(
            os.path.dirname(__file__), "runtime_environment", BOOTSTRAP_SCRIPT_NAME
        )
        runtime_manager_script_path = os.path.join(
            os.path.dirname(__file__), "runtime_environment", RUNTIME_MANAGER_SCRIPT_NAME
        )

        # copy runtime scripts to tmpdir
        shutil.copy2(bootstrap_script_path, bootstrap_scripts)
        shutil.copy2(runtime_manager_script_path, bootstrap_scripts)

        """"
        upload_path = S3Uploader.upload(
            bootstrap_scripts,
            s3_path_join(s3_base_uri, RUNTIME_SCRIPTS_CHANNEL_NAME),
            s3_kms_key,
            sagemaker_session,
        )
        return upload_path
        """
        return None
        

def _prepare_deps_and_pre_execution_scripts(
    local_deps_path: str,
    pre_execution_commands: List[str],
    pre_execution_script_local_path: str,
    tmp_dir: str,
):
    """Prepare pre-execution scripts and deps and upload them to s3.

    If pre execution commands are provided, a new bash file will be created
      with those commands in tmp directory.
    If pre execution script is provided, it copies that file from local file path
      to tmp directory.
    If local deps file is provided, it copies that file from local file path
      to tmp directory.
    If under pipeline context, tmp directory with copied deps and scripts is
      uploaded to S3.
    """

    if not (local_deps_path or pre_execution_commands or pre_execution_script_local_path):
        return None

    if local_deps_path:
        dst_path = shutil.copy2(local_deps_path, tmp_dir)
        # workspace_file_tree.add(f"[yellow]{os.path.basename(local_deps_path)}, size: {os.path.getsize(local_deps_path)} bytes[/yellow]")
        logger.info("Copied deps file at '%s' to '%s'", local_deps_path, dst_path)

    if pre_execution_commands or pre_execution_script_local_path:
        pre_execution_script = os.path.join(tmp_dir, PRE_EXECUTION_SCRIPT_NAME)
        if pre_execution_commands:
            with open(pre_execution_script, "w") as target_script:
                commands = [cmd + "\n" for cmd in pre_execution_commands]
                target_script.writelines(commands)
                logger.info(
                    "Generated pre-execution script from commands to '%s'", pre_execution_script
                )
        else:
            shutil.copy2(pre_execution_script_local_path, pre_execution_script)
            logger.info(
                "Copied pre-execution commands from script at '%s' to '%s'",
                pre_execution_script_local_path,
                pre_execution_script,
            )
    return None

def _prepare_and_upload_workspace(
    job_id: str,
    local_deps_path: str,
    include_local_workdir: bool,
    pre_execution_commands: List[str],
    pre_execution_script_local_path: str,
    custom_file_filter: Optional[Union[Callable[[str, List], List], CustomFileFilter]] = None,
) -> str:
    """Prepare and upload the workspace to S3. """

    if not (
        local_deps_path
        or include_local_workdir
        or pre_execution_commands
        or pre_execution_script_local_path
    ):
        return None

    with _tmpdir() as tmp_dir:
        tmp_workspace_dir = os.path.join(tmp_dir, "temp_workspace/")
        os.mkdir(tmp_workspace_dir)
        
        # TODO Remove the following hack to avoid dir_exists error in the copy_tree call below.
        # UPDATE: fixed with the dirs_exist_ok=True flag in the copy_tree call below.
        # tmp_workspace = os.path.join(tmp_workspace_dir, JOB_REMOTE_FUNCTION_WORKSPACE)

        tmp_workspace = tmp_workspace_dir

        if include_local_workdir:
            copy_workdir(
                dst=tmp_workspace, 
                custom_file_filter=custom_file_filter,
                print_tree=False)
            logger.info(f"Copied local directory {os.getcwd()} to '%s'", tmp_workspace)

        if not os.path.isdir(tmp_workspace):
            # create the directory if no workdir_path was provided in the input.
            os.mkdir(tmp_workspace)

        _prepare_deps_and_pre_execution_scripts(
            local_deps_path=local_deps_path,
            pre_execution_commands=pre_execution_commands,
            pre_execution_script_local_path=pre_execution_script_local_path,
            tmp_dir=tmp_workspace,
        )

        workspace_archive_path = os.path.join(tmp_dir, "workspace")
        workspace_archive_path = shutil.make_archive(
            workspace_archive_path, 
            "zip", 
            tmp_workspace_dir
        )
        logger.info("Successfully created workdir archive at '%s'", workspace_archive_path)

        upload_path = S3Uploader.upload_file(
            workspace_archive_path,
            job_id,
            "workspace.zip"
        )

        logger.info("Successfully uploaded workdir to '%s'", upload_path)
        return upload_path


